/*
mediastreamer2 library - modular sound and video processing and streaming
Copyright (C) 2006  Simon MORLAT (simon.morlat@linphone.org)
Copyright (C) 2012  Aymeric Moizard (amoizard@gmail.com)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "mediastreamer2/msplc.h"
#ifdef HAVE_CONFIG_H
#include "mediastreamer-config.h"
#endif

static MSPlcDesc ms_plc_desc = {
  0,
  "",
  0xFFFF,
	NULL,
	NULL,
	NULL,
	NULL,
  NULL,
};

void ms_plc_set_func(MSPlcDesc *_ms_plc_desc)
{
  if (_ms_plc_desc->struct_size!=sizeof(MSPlcDesc))
    return;
	if (_ms_plc_desc->quality_priority<=ms_plc_desc.quality_priority || ms_plc_desc.plc_init==NULL)
	{
    /* faster plugin detected: replace all method */
    memcpy(&ms_plc_desc, _ms_plc_desc, sizeof(MSPlcDesc));
  	ms_message("msplc: plugin for PLC registered: %s", ms_plc_desc.name);
		return;
	}
}

MSPlc *ms_plc_init(int samplerate)
{
  MSPlc *obj;
  int res;
  if (ms_plc_desc.plc_init==NULL)
    return NULL;

  obj = (MSPlc *)ms_new0(MSPlc,1);
  obj->samplerate = samplerate;
  res = ms_plc_desc.plc_init(obj);
  if (res<0) {
    ms_free(obj);
    return NULL;
  }
  return obj;
}

void ms_plc_free(MSPlc *obj)
{
  if (ms_plc_desc.plc_init==NULL)
    return;
  if (obj==NULL)
    return;
  ms_plc_desc.plc_free(obj);
  ms_free(obj);
}

int ms_plc_expand(MSPlc *obj, mblk_t  **out, int length, int noise)
{
  if (ms_plc_desc.plc_init==NULL)
    return -1;
  if (obj==NULL)
    return -1;
  return ms_plc_desc.plc_expand(obj, out, length, noise);
}

int ms_plc_normal(MSPlc *obj, mblk_t  *in, mblk_t  **out)
{
  if (ms_plc_desc.plc_init==NULL)
    return -1;
  if (obj==NULL)
    return -1;
  return ms_plc_desc.plc_normal(obj, in, out);
}

int ms_plc_reduce(MSPlc *obj, mblk_t  *in, mblk_t  **out)
{
  if (ms_plc_desc.plc_init==NULL)
    return -1;
  if (obj==NULL)
    return -1;
  return ms_plc_desc.plc_reduce(obj, in, out);
}
