#if 0 //QTkit deprecated with xcode 8 - use newer AVFoudation msv4m.m

#include "mediastreamer-config.h"
#include "mediastreamer2/msvideo.h"
#include "mediastreamer2/msticker.h"
#include "mediastreamer2/msv4l.h"
#include "mediastreamer2/mswebcam.h"
#include "nowebcam.h"

#import <QTKit/QTkit.h>

#define ENABLE_FILTER_STATISTICS

struct v4mState;

@interface NsMsWebCam :NSObject
{
	QTCaptureDeviceInput *input;
	QTCaptureDecompressedVideoOutput * output;
	QTCaptureSession *session;
	//QTCaptureDevice *device;
	
	ms_mutex_t mutex;
	queue_t rq;
};

-(id) init;
-(void) dealloc;
-(int) start;
-(int) stop;
-(void) setSize:(MSVideoSize) size;
-(MSVideoSize) getSize;
-(void) setName:(char*) name;
-(int) convertToMSPixFmt:(UInt32)fmt;
-(int) getPixFmt;

-(QTCaptureSession *) session;
-(queue_t*) rq;
-(ms_mutex_t *) mutex;

@end

@implementation NsMsWebCam

-(void)captureOutput:(QTCaptureOutput *)captureOutput didOutputVideoFrame:(CVImageBufferRef)videoFrame withSampleBuffer:(QTSampleBuffer *)sampleBuffer fromConnection:
(QTCaptureConnection *)connection
{
  mblk_t *buf;
	ms_mutex_lock(&mutex);
  
  CVReturn status = CVPixelBufferLockBaseAddress(videoFrame, 0);
  if (kCVReturnSuccess != status) {
    ms_mutex_unlock(&mutex);
    return;
  }
  
  UInt32 fmt  = CVPixelBufferGetPixelFormatType(videoFrame);
  int format = [self convertToMSPixFmt:fmt];
  if (format<0) {
    CVPixelBufferUnlockBaseAddress(videoFrame, 0);
    ms_mutex_unlock(&mutex);
    return;
  }
  int w = CVPixelBufferGetWidth(videoFrame);
  int h = CVPixelBufferGetHeight(videoFrame);
  
	if (CVPixelBufferIsPlanar(videoFrame)) {
		size_t nb_planes = CVPixelBufferGetPlaneCount(videoFrame);
		MSPicture pict;
		buf = yuv_buf_alloc(&pict, w, h);
		int p;
    
		for (p=0; p < nb_planes; p++) {
			size_t byteperrow = CVPixelBufferGetBytesPerRowOfPlane(videoFrame, p);
			size_t width_of_plane = CVPixelBufferGetWidthOfPlane(videoFrame, p);
			size_t height_of_plane = CVPixelBufferGetHeightOfPlane(videoFrame, p);
			uint8_t *dst_plane = pict.planes[p];
			uint8_t *src_plane = CVPixelBufferGetBaseAddressOfPlane(videoFrame, p);
			int l;
			for (l=0; l<height_of_plane; l++) {
				memcpy(dst_plane, src_plane, width_of_plane);
				src_plane += byteperrow;
				dst_plane += width_of_plane;
			}
		}
	} else {
    size_t data_size = CVPixelBufferGetDataSize(videoFrame);
    uint8_t *src_buf = CVPixelBufferGetBaseAddress(videoFrame);
    buf=allocb(data_size,63);
    memcpy(buf->b_wptr, src_buf, data_size);
    buf->b_wptr+=data_size;
  }
  
  CVPixelBufferUnlockBaseAddress(videoFrame, 0);
  
  mblk_set_video_format(buf, format);
  mblk_set_video_width(buf, w);
  mblk_set_video_height(buf, h);
  putq(&rq, buf);
  
	ms_mutex_unlock(&mutex);
}

-(id) init {
	qinit(&rq);
	ms_mutex_init(&mutex,NULL);
	
	session = [[QTCaptureSession alloc] init];
	output = [[QTCaptureDecompressedVideoOutput alloc] init];
	
	[output setAutomaticallyDropsLateVideoFrames:YES];
	
	[output setDelegate: self];
	
	
	return self;
}

-(void) dealloc {
	[self stop];
	
	if(session)
	{
		[session release];
		session = nil;
	}
  
	if(input)
	{
		[input release];
		input = nil;
	}
	
	if(output)
	{
		[output release];
		output = nil;
	}
	
	flushq(&rq,0);
	ms_mutex_destroy(&mutex);
  
	[super dealloc];
}

-(int) start {
	
	[session startRunning];
	
	return 0;
}

-(int) stop {
	
	[session stopRunning];
	
	return 0;
}

-(int) convertToMSPixFmt:(UInt32)fmt{
  
  if( fmt == k2vuyPixelFormat) //k422YpCbCr8CodecType)
  {
    return MS_UYVY;
  }
  else if( fmt == kCVPixelFormatType_420YpCbCr8Planar)
  {
    return MS_YUV420P;
  }
  
  else if( fmt == kCVPixelFormatType_24RGB)
  {
    return MS_RGB24;
  }
  else if( fmt == kUYVY422PixelFormat)
  {
    return MS_UYVY;
  }
  else if( fmt == kYUVSPixelFormat)
  {
    return MS_YUY2;
  }
  else if( fmt == k32RGBAPixelFormat)
  {
    return MS_RGBA32;
  }
  return -1;
}

-(int) getPixFmt{
  QTCaptureDevice *device = [input device];
	if([device isOpen])
	{
		NSArray * array = [device formatDescriptions];
    
		NSEnumerator *enumerator = [array objectEnumerator];
		QTFormatDescription *desc;
		while ((desc = [enumerator nextObject]))
		{
			if ([desc mediaType] == QTMediaTypeVideo)
			{
				UInt32 fmt = [desc formatType];
				
				if (fmt > 0x28)
					NSLog(@"format for webcam: %c%c%c%c", (char)((fmt >> 24) & 0xFF),
                (char)((fmt >> 16) & 0xFF),
                (char)((fmt >>  8) & 0xFF),
                (char)((fmt >>  0) & 0xFF));
				else
					NSLog(@"format for webcam: %d", (unsigned int)fmt);
        
        int format = [self convertToMSPixFmt:fmt];
        if (format>=0)
          return format;
      }
    }
		NSLog(@"qtcapturedevice no format detected");
		NSDictionary * dic = [NSDictionary dictionaryWithObjectsAndKeys:
                          [NSNumber numberWithInteger:k32RGBAPixelFormat], (id)kCVPixelBufferPixelFormatTypeKey,
                          nil];
		[output setPixelBufferAttributes:dic];
		return MS_YUV420P;
  }
	
	NSLog(@"qtcapturedevice fallback to MS_YUV420P");
	return MS_YUV420P;
}

-(void) setName:(char*) name
{
	NSError *error = nil;
	unsigned int i = 0;
  
	QTCaptureDevice * device=nil;
	
	if(name != nil)
	{
		NSLog(@"card name = %s", name);
		NSArray * array = [QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeVideo];
    
		for(i = 0 ; i < [array count]; i++)
		{
			QTCaptureDevice * deviceTmp = [array objectAtIndex:i];
			NSLog(@"device name = %s", [[deviceTmp localizedDisplayName] UTF8String]);
			if(strcmp([[deviceTmp localizedDisplayName] UTF8String], name)==0)
			{
				device = deviceTmp;
				break;
			}
		}
	}
	
	if(device==nil)
		device = [QTCaptureDevice defaultInputDeviceWithMediaType:QTMediaTypeVideo];
	if(device==nil)
		return;
  
	if(![device open:&error])
		return;
	
	input = [[QTCaptureDeviceInput alloc] initWithDevice:device];
	
	[session addInput:input error:&error];
	[session addOutput:output error:&error];
  
}

-(void) setSize:(MSVideoSize) size
{
	NSDictionary * dic = [NSDictionary dictionaryWithObjectsAndKeys:
                        [NSNumber numberWithInteger:size.width], (id)kCVPixelBufferWidthKey,
                        [NSNumber numberWithInteger:size.height],(id)kCVPixelBufferHeightKey,
                        //[NSNumber numberWithInteger:kCVPixelFormatType_420YpCbCr8Planar], (id)kCVPixelBufferPixelFormatTypeKey,
                        nil];
	
	[output setPixelBufferAttributes:dic];
}

-(MSVideoSize) getSize
{
	MSVideoSize size;
	
	size.width = MS_VIDEO_SIZE_QCIF_W;
	size.height = MS_VIDEO_SIZE_QCIF_H;
	
	if(output)
	{
		NSDictionary * dic = [output pixelBufferAttributes];
		
		size.width = (int)[[dic objectForKey:(id)kCVPixelBufferWidthKey] integerValue];
		size.height = (int)[[dic objectForKey:(id)kCVPixelBufferHeightKey] integerValue];
	}
	
	return size;
}

-(QTCaptureSession *) session
{
	return	session;
}

-(queue_t*) rq
{
	return &rq;
}

-(ms_mutex_t *) mutex
{
	return &mutex;
}

@end

typedef struct v4mState{
	NsMsWebCam * webcam;
	NSAutoreleasePool* myPool;
	mblk_t *mire;
	int frame_ind;
	float fps;
	float start_time;
	int frame_count;
	bool_t usemire;
  
	MSFilterFps *stats;
}v4mState;

static void v4m_init(MSFilter *f){
	v4mState *s=ms_new0(v4mState,1);
	//s->myPool = [[NSAutoreleasePool alloc] init];
	s->webcam= [[NsMsWebCam alloc] init];
	//[s->webcam retain];
	s->mire=NULL;
	s->start_time=0;
	s->frame_count=-1;
	s->fps=15;
	s->usemire=(getenv("DEBUG")!=NULL);
	s->stats=NULL;
#ifdef ENABLE_FILTER_STATISTICS
	s->stats=ms_fps_init(2000);
#endif
	f->data=s;
}

static int v4m_start(MSFilter *f, void *arg) {
	v4mState *s=(v4mState*)f->data;
	[s->webcam start];
	ms_message("v4m video device opened.");
	return 0;
}

static int v4m_stop(MSFilter *f, void *arg){
	v4mState *s=(v4mState*)f->data;
	[s->webcam stop];
	ms_message("v4m video device closed.");
	
	return 0;
}

static void v4m_uninit(MSFilter *f){
	v4mState *s=(v4mState*)f->data;
	v4m_stop(f,NULL);
	
	freemsg(s->mire);
	if (s->stats) ms_free(s->stats);
	[s->webcam release];
	//[s->myPool release];
	ms_free(s);
}

static mblk_t * v4m_make_mire(v4mState *s){
	unsigned char *data;
	int i,j,line,pos;
	MSVideoSize vsize = [s->webcam getSize];
	int patternw=vsize.width/6;
	int patternh=vsize.height/6;
	int red,green=0,blue=0;
	if (s->mire==NULL){
		s->mire=allocb(vsize.width*vsize.height*3,63);
		s->mire->b_wptr=s->mire->b_datap->db_lim;
	}
	data=s->mire->b_rptr;
	for (i=0;i<vsize.height;++i){
		line=i*vsize.width*3;
		if ( ((i+s->frame_ind)/patternh) & 0x1) red=255;
		else red= 0;
		for (j=0;j<vsize.width;++j){
			pos=line+(j*3);
			
			if ( ((j+s->frame_ind)/patternw) & 0x1) blue=255;
			else blue= 0;
			
			data[pos]=red;
			data[pos+1]=green;
			data[pos+2]=blue;
		}
	}
	s->frame_ind++;
	return s->mire;
}

static mblk_t * v4m_make_nowebcam(v4mState *s){
	if (s->mire==NULL && s->frame_ind==0){
		//s->mire=ms_load_nowebcam(&s->vsize, -1);
	}
	s->frame_ind++;
	return s->mire;
}

static void v4m_process(MSFilter * obj){
	v4mState *s=(v4mState*)obj->data;
	uint32_t timestamp;
	int cur_frame;
	if (s->frame_count==-1){
		s->start_time=obj->ticker->time;
		s->frame_count=0;
	}
  
	ms_mutex_lock([s->webcam mutex]);
  
	cur_frame=((obj->ticker->time-s->start_time)*s->fps/1000.0);
	if (cur_frame>=s->frame_count)
	{
		mblk_t *om=NULL;
		/*keep the most recent frame if several frames have been captured */
		if ([[s->webcam session] isRunning])
		{
			om=getq([s->webcam rq]);
		}
		else
		{
			/*if (s->pix_fmt==MS_YUV420P && s->vsize.width==MS_VIDEO_SIZE_CIF_W && s->vsize.height==MS_VIDEO_SIZE_CIF_H)
       {
       if (s->usemire)
       {
       om=dupmsg(v4m_make_mire(s));
       }
       else
       {
       mblk_t *tmpm=v4m_make_nowebcam(s);
       if (tmpm)
       om=dupmsg(tmpm);
       }
       }*/
		}
		
		if (om!=NULL)
		{
			timestamp=(uint32_t)(obj->ticker->time*90);/* rtp uses a 90000 Hz clockrate for video*/
			mblk_set_timestamp_info(om,timestamp);
			mblk_set_marker_info(om,TRUE);
			ms_fps_process(s->stats, obj, obj->ticker->time);
			ms_queue_put(obj->outputs[0],om);
			s->frame_count++;
		}
		if ([s->webcam rq]->q_mcount>1)
			flushq([s->webcam rq],0);
	}
	else
		flushq([s->webcam rq],0);
  
	ms_mutex_unlock([s->webcam mutex]);
}

static void v4m_preprocess(MSFilter *f){
	v4m_start(f,NULL);
  
}

static void v4m_postprocess(MSFilter *f){
	v4m_stop(f,NULL);
}

static int v4m_set_fps(MSFilter *f, void *arg){
	v4mState *s=(v4mState*)f->data;
	s->fps=*((float*)arg);
	s->frame_count=-1;
	return 0;
}

static int v4m_get_pix_fmt(MSFilter *f,void *arg){
	v4mState *s=(v4mState*)f->data;
	*((MSPixFmt*)arg) = [s->webcam getPixFmt];
	return 0;
}

static int v4m_set_vsize(MSFilter *f, void *arg){
	v4mState *s=(v4mState*)f->data;
	[s->webcam setSize:*((MSVideoSize*)arg)];
	return 0;
}

static int v4m_get_vsize(MSFilter *f, void *arg){
	v4mState *s=(v4mState*)f->data;
	*(MSVideoSize*)arg = [s->webcam getSize];
	return 0;
}

static MSFilterMethod methods[]={
	{	MS_FILTER_SET_FPS		,	v4m_set_fps		},
	{	MS_FILTER_GET_PIX_FMT	,	v4m_get_pix_fmt	},
	{	MS_FILTER_SET_VIDEO_SIZE, 	v4m_set_vsize	},
	{	MS_V4L_START			,	v4m_start		},
	{	MS_V4L_STOP				,	v4m_stop		},
	{	MS_FILTER_GET_VIDEO_SIZE,	v4m_get_vsize	},
	{	0						,	NULL			}
};

MSFilterDesc ms_v4m_desc={
	.id=MS_V4L_ID,
	.name="MSV4m",
	.text="A video for macosx compatible source filter to stream pictures.",
	.ninputs=0,
	.noutputs=1,
	.category=MS_FILTER_OTHER,
	.init=v4m_init,
	.preprocess=v4m_preprocess,
	.process=v4m_process,
	.postprocess=v4m_postprocess,
	.uninit=v4m_uninit,
	.methods=methods
};

MS_FILTER_DESC_EXPORT(ms_v4m_desc)

static void ms_v4m_detect(MSWebCamManager *obj);

static void ms_v4m_cam_init(MSWebCam *cam)
{
}

static int v4m_set_device(MSFilter *f, void *arg)
{
	/*
   v4mState *s=(v4mState*)f->data;
   s->id = (char*) malloc(sizeof(char)*strlen((char*)arg));
   strcpy(s->id,(char*)arg);
	 */
  
	return 0;
}

static int v4m_set_name(MSFilter *f, void *arg){
	
	v4mState *s=(v4mState*)f->data;
  
	//s->name = (char*) malloc(sizeof(char)*strlen((char*)arg));
	//strcpy(s->name,(char*)arg);
	
	[s->webcam setName:(char*)arg];
  
	return 0;
}

static MSFilter *ms_v4m_create_reader(MSWebCam *obj)
{
	MSFilter *f= ms_filter_new_from_desc(&ms_v4m_desc);
  
	v4m_set_device(f,obj->id);
	v4m_set_name(f,obj->name);
  
	return f;
}

MSWebCamDesc ms_v4m_cam_desc={
	"VideoForMac grabber",
	&ms_v4m_detect,
  NULL,
	&ms_v4m_cam_init,
	&ms_v4m_create_reader,
	NULL
};


static void ms_v4m_detect(MSWebCamManager *obj){
  
	unsigned int i = 0;
	NSAutoreleasePool* myPool = [[NSAutoreleasePool alloc] init];
	
	NSArray * array = [QTCaptureDevice inputDevicesWithMediaType:QTMediaTypeVideo];
	
	for(i = 0 ; i < [array count]; i++)
	{
		QTCaptureDevice * device = [array objectAtIndex:i];
		MSWebCam *cam=ms_web_cam_new(&ms_v4m_cam_desc);
		
		cam->name= ms_strdup([[device localizedDisplayName] UTF8String]);
		cam->id = ms_strdup([[device uniqueID] UTF8String]);
		cam->data = NULL;
		ms_web_cam_manager_add_cam(obj,cam);
		
		NSArray *format_array = [device formatDescriptions];
		QTFormatDescription* camera_format = NULL;
		for( int k=0; k < [format_array count]; k++ )
		{
			camera_format = [format_array objectAtIndex: k];
			
			NSLog(@"%@", [camera_format localizedFormatSummary] );
			NSLog(@"%@", [[camera_format formatDescriptionAttributes] description] );
		}
	}
	[myPool release];
}

#endif
