/*
mediastreamer2 library - modular sound and video processing and streaming
Copyright (C) 2006  Simon MORLAT (simon.morlat@linphone.org)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#include "mediastreamer-config.h"
#endif

#define NOWINDOW_DISPLAY

#ifdef __APPLE__
#include "TargetConditionals.h"
#undef HAVE_SDL

#if (TARGET_OS_IPHONE==0) && (TARGET_IPHONE_SIMULATOR==0)
//ADD THIS DEFINE TO DISPLAY PIN 0 and PIN 1 in ONE VIEW
//#define NOWINDOW_DISPLAY_ONEVIEW
#endif
#if (TARGET_OS_IPHONE==0) && (TARGET_IPHONE_SIMULATOR==0)
//ADD THIS DEFINE TO RETREIVE PIN 0 and PIN 1 in YUY2 format
#define NOWINDOW_DISPLAY_UYVY
#endif

#elif defined(__linux)
#define NOWINDOW_DISPLAY_YUV420P
#endif



#include "mediastreamer2/msfilter.h"
#include "mediastreamer2/msvideo.h"

/*required for dllexport of win_display_desc */
#define INVIDEOUT_C 1
#include "mediastreamer2/msvideoout.h"

#ifdef VIDEO_ENABLED

#define SCALE_FACTOR 4.0f
#define SELVIEW_POS_INACTIVE -100.0

static int video_out_set_vsize(MSFilter *f,void *arg);

int ms_display_poll_event(MSDisplay *d, MSDisplayEvent *ev){
  if (d->desc->pollevent)
    return d->desc->pollevent(d,ev);
  else return -1;
}

//#undef HAVE_X11_EXTENSIONS_XV_H
#if defined(HAVE_X11_EXTENSIONS_XV_H) && defined(HAVE_X11_XLIB_H)

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>

typedef struct XvDisplay {
  MSFilter *filter;
  ms_mutex_t xv_mutex;
  Display *display;
  Window window;  
  GC gc;
  XImage* ximage;  
  XvImage *xvImage;
  int XvPort;
  int XvFormat;
  float sv_scalefactor;
  MSVideoSize screen_size;

  MSPicture fb;
  mblk_t *fb_block;
  MSPicture fb_selfview;
  mblk_t *fb_selfview_block;
  
  XWindowChanges values_position;
} XvDisplay;

static long xv_get_native_window_id(XvDisplay *wd){
  return (long) wd->window;
}

static void xv_show_window(XvDisplay *wd, bool_t show){
  if (wd->display==NULL || wd->window<=0)
    return;

  if (show)
    XMapWindow(wd->display,wd->window);
  else
    XUnmapWindow(wd->display,wd->window);
}


static void xv_display_uninit(MSDisplay *obj);

static int xv_create_window(XvDisplay *wd, int w, int h){
  //static bool_t once=TRUE;
  XvAdaptorInfo *xv_adaptor_info=NULL;
  XvAdaptorInfo *ai;
  unsigned int xv_adaptors;
  unsigned int i;
  XGCValues xv_xgc;

  int screen;
  XVisualInfo vinfo;
  XWindowAttributes wattr;
  unsigned long xswamask;
  XSetWindowAttributes xswa;
  XSizeHints hint;

#if 0
  Atom allowed_action;
  Atom allowed_actions[7];
#endif
  
  hint.x=50;
  hint.y=50;
  hint.width=w;
  hint.height=h;
  hint.min_width=MS_VIDEO_SIZE_CIF_W;
  hint.min_height=MS_VIDEO_SIZE_CIF_H;
  hint.max_width=MS_VIDEO_SIZE_720P_W;
  hint.max_height=MS_VIDEO_SIZE_720P_H;
  hint.flags=PPosition | PSize | PMinSize | PMaxSize;

  //wd->display = XOpenDisplay(":0.0");
  wd->display = XOpenDisplay(NULL);
  if (wd->display==NULL){
    ms_error("videout.c:xv XOpenDisplay failed");
    return -1;
  }
  ms_message("videout.c:xv XOpenDisplay done");

#if 0
  allowed_action = XInternAtom(wd->display, "_NET_WM_ALLOWED_ACTIONS", True);
  allowed_actions[0] = XInternAtom(wd->display, "_NET_WM_ACTION_MOVE", True);
  allowed_actions[1] = XInternAtom(wd->display, "_NET_WM_ACTION_RESIZE", True);
  allowed_actions[2] = XInternAtom(wd->display, "_NET_WM_ACTION_MAXIMIZE_HORZ", False);
  allowed_actions[3] = XInternAtom(wd->display, "_NET_WM_ACTION_MAXIMIZE_VERT", False);
  allowed_actions[4] = XInternAtom(wd->display, "_NET_WM_ACTION_FULLSCREEN", False);
  allowed_actions[5] = XInternAtom(wd->display, "_NET_WM_ACTION_CHANGE_DESKTOP", True);
  allowed_actions[6] = XInternAtom(wd->display, "_NET_WM_ACTION_CLOSE", False);
#endif

  XGetWindowAttributes(wd->display, DefaultRootWindow(wd->display), &wattr);
  screen = DefaultScreen(wd->display);
  XMatchVisualInfo(wd->display, screen, wattr.depth, TrueColor, &vinfo);

  ms_message("videout.c:xv depth=%i", wattr.depth);
  xswa.override_redirect = 1;
  //xswa.override_redirect = 0;
  xswa.background_pixel = 0;
  xswa.border_pixel = 10;
  xswamask = CWBackPixel | CWBorderPixel | CWOverrideRedirect;
  wd->window = XCreateWindow(wd->display,
    RootWindow(wd->display, screen),
    hint.x, hint.y, hint.width, hint.height, 4,
    wattr.depth,
    CopyFromParent,
    vinfo.visual,
    xswamask,
    &xswa);
  if (wd->window<=0){
    if (wd->display!=NULL)
      XCloseDisplay(wd->display);
    wd->display=NULL;
    ms_error("videout.c:xv XCreateWindow failed");
    return -1;
  }

  {
    Atom modal;
    Atom state;
    state = XInternAtom(wd->display, "_NET_WM_STATE", True);
    modal = XInternAtom(wd->display, "_NET_WM_STATE_MODAL", True);
    XChangeProperty(wd->display, wd->window, state, XA_ATOM, 32, PropModeReplace, (unsigned char*)&modal, 1);
  }
#if 0
  XChangeProperty(wd->display, wd->window, allowed_action, XA_ATOM, 32, PropModeReplace, (unsigned char*)&allowed_actions, 7);
#endif

  {
    Atom window_type = XInternAtom (wd->display, "_NET_WM_WINDOW_TYPE", True);
    Atom type_dock = XInternAtom (wd->display, "_NET_WM_WINDOW_TYPE_SPLASH", True);
    //Atom type_dock = XInternAtom (wd->display, "_NET_WM_WINDOW_TYPE_TOOLBAR", True); //only a close button

    XChangeProperty (wd->display, wd->window, window_type, XA_ATOM, 32, PropModeReplace, (unsigned char *)&type_dock, 1);
  }


  xv_xgc.graphics_exposures = 0;
  wd->gc = XCreateGC(wd->display, wd->window, 0L, &xv_xgc);
  if (wd->gc==NULL){
    ms_error("videout.c:xv XCreateGC failed");
    if (wd->window>0)
      XDestroyWindow(wd->display, wd->window);
    wd->window=0;
    if (wd->display!=NULL)
      XCloseDisplay(wd->display);
    wd->display=NULL;
    return -1;
  }

  if (Success != XvQueryAdaptors(wd->display,
    wd->window,
    &xv_adaptors, 
    (XvAdaptorInfo **)&xv_adaptor_info)) {
  }

  ai = (XvAdaptorInfo *)xv_adaptor_info;

  wd->XvPort=0;
  for (i = 0; i < xv_adaptors; i++) {
    ms_message("adaptors: name: %s baseport=%i", ai[i].name, ai[i].base_id);
    if ((ai[i].type &XvInputMask) && (ai[i].type &XvImageMask))
    {
      XvPortID xv_p;
      for (xv_p = ai[i].base_id;
        xv_p < ai[i].base_id + ai[i].num_ports; ++xv_p){
          if (XvGrabPort(wd->display, xv_p, CurrentTime)==0)
          {
            wd->XvPort = xv_p;
            break;
          }
      }
      if (wd->XvPort!=0)
        break;
    }
  }
  XvFreeAdaptorInfo(ai);
  ms_message("using Xvideo port %d for hw scaling", wd->XvPort);
  if (wd->XvPort>0)
  {
    uint32_t fmt = 0x32315659; //YV12
    int formats;
    XvImageFormatValues *fo;
    fo = XvListImageFormats(wd->display, wd->XvPort, (int *)&formats);
    for (i=0;i<formats;i++)
    {
      ms_message("format=[%4.4s] %s", (char *) &fo[i].id, (fo[i].format == XvPacked) ? "packed" : "planar");
      if (fo[i].id==fmt){
        wd->XvFormat =fo[i].id;
        ms_message("found YV12 format");
      }
    }
    XFree(fo);
  }

  wd->xvImage = (XvImage *)XvCreateImage( wd->display , wd->XvPort, wd->XvFormat, NULL, w, h);
  return 0;
}

static bool_t xv_display_init(MSDisplay *obj, MSFilter *f, MSPicture *fbuf, MSPicture *fbuf_selfview){
  XvDisplay *wd = (XvDisplay*)obj->data;
  int i;

  if (wd==NULL){
    //char driver[128];
    /* Initialize the SDL library */
    wd=(XvDisplay*)ms_new0(XvDisplay,1);
    wd->filter = f;
    obj->data=wd;

    wd->values_position.x = 50;
    wd->values_position.y = 50;
    
    ms_mutex_init(&wd->xv_mutex,NULL);
    ms_mutex_lock(&wd->xv_mutex);
    wd->screen_size.width = fbuf->w;
    wd->screen_size.height = fbuf->h;

    wd->XvFormat=0x32315659; //IMGFMT_YV12;

    i=xv_create_window(wd, fbuf->w, fbuf->h);
    if (i==0)
      xv_show_window(wd, TRUE);

    if (wd->xvImage!=NULL) {
      wd->fb_block = allocb(wd->xvImage->data_size, 63);
      wd->fb.planes[0]= wd->fb_block->b_rptr + wd->xvImage->offsets[0];
      wd->fb.planes[1]= wd->fb_block->b_rptr + wd->xvImage->offsets[1];
      wd->fb.planes[2]= wd->fb_block->b_rptr + wd->xvImage->offsets[2];
      wd->fb.planes[3]= 0;
      wd->fb.strides[0]= wd->xvImage->pitches[0];
      wd->fb.strides[1]= wd->xvImage->pitches[1];
      wd->fb.strides[2]= wd->xvImage->pitches[2];
      wd->fb.strides[3]= 0;
      wd->fb.w=fbuf->w;
      wd->fb.h=fbuf->h;
      wd->fb_block->b_wptr = wd->fb_block->b_rptr+wd->xvImage->data_size;
      wd->xvImage->data = wd->fb_block->b_rptr;
    }
  }else {
    ms_mutex_lock(&wd->xv_mutex);

    if (wd->display==NULL) {

      if (wd->fb_block!=NULL)
	freemsg(wd->fb_block);
      wd->fb_block=NULL;

      i=xv_create_window(wd, fbuf->w, fbuf->h);
      if (i==0)
	xv_show_window(wd, TRUE);

      if (wd->xvImage!=NULL) {
	wd->fb_block = allocb(wd->xvImage->data_size, 63);
	wd->fb.planes[0]= wd->fb_block->b_rptr + wd->xvImage->offsets[0];
	wd->fb.planes[1]= wd->fb_block->b_rptr + wd->xvImage->offsets[1];
	wd->fb.planes[2]= wd->fb_block->b_rptr + wd->xvImage->offsets[2];
	wd->fb.planes[3]= 0;
	wd->fb.strides[0]= wd->xvImage->pitches[0];
	wd->fb.strides[1]= wd->xvImage->pitches[1];
	wd->fb.strides[2]= wd->xvImage->pitches[2];
	wd->fb.strides[3]= 0;
	wd->fb.w=fbuf->w;
	wd->fb.h=fbuf->h;
	wd->fb_block->b_wptr = wd->fb_block->b_rptr+wd->xvImage->data_size;
	wd->xvImage->data = wd->fb_block->b_rptr;
      }
    } else {
      if (wd->screen_size.width!=fbuf->w || wd->screen_size.height != fbuf->h || wd->xvImage->data != wd->fb_block->b_rptr) {
	
	i = XResizeWindow(wd->display, wd->window, fbuf->w, fbuf->h);
	if (i<=0) {
	  ms_message("xv window not resized(err=%i) to: %i, %i", i, fbuf->w, fbuf->h);
	} else {
	  ms_message("xv window resized to: %i, %i", fbuf->w, fbuf->h);
	}
	wd->screen_size.width = fbuf->w;
	wd->screen_size.height = fbuf->h;
	
	if (wd->xvImage!=NULL)
	  XFree(wd->xvImage);
	wd->xvImage=NULL;
	if (wd->fb_block!=NULL)
	  freemsg(wd->fb_block);
	wd->fb_block=NULL;

	wd->xvImage = (XvImage *)XvCreateImage( wd->display , wd->XvPort, wd->XvFormat, NULL, wd->screen_size.width, wd->screen_size.height);

	if (wd->xvImage!=NULL) {
	  wd->fb_block = allocb(wd->xvImage->data_size, 63);
	  wd->fb.planes[0]= wd->fb_block->b_rptr + wd->xvImage->offsets[0];
	  wd->fb.planes[1]= wd->fb_block->b_rptr + wd->xvImage->offsets[1];
	  wd->fb.planes[2]= wd->fb_block->b_rptr + wd->xvImage->offsets[2];
	  wd->fb.planes[3]= 0;
	  wd->fb.strides[0]= wd->xvImage->pitches[0];
	  wd->fb.strides[1]= wd->xvImage->pitches[1];
	  wd->fb.strides[2]= wd->xvImage->pitches[2];
	  wd->fb.strides[3]= 0;
	  wd->fb.w=fbuf->w;
	  wd->fb.h=fbuf->h;
	  wd->fb_block->b_wptr = wd->fb_block->b_rptr+wd->xvImage->data_size;
	  wd->xvImage->data = wd->fb_block->b_rptr;
	}
      }
      i=0;
    }
  }

  wd->filter = f;


  if (wd->display!=NULL) {
    XRaiseWindow(wd->display, wd->window);
    obj->window_id=xv_get_native_window_id(wd);
  }
  if (wd->fb_block!=NULL){

    fbuf->planes[0]=wd->fb.planes[0];
    fbuf->planes[1]=wd->fb.planes[2];
    fbuf->planes[2]=wd->fb.planes[1];
    fbuf->planes[3]=NULL;
    fbuf->strides[0]=wd->fb.strides[0];
    fbuf->strides[1]=wd->fb.strides[2];
    fbuf->strides[2]=wd->fb.strides[1];
    fbuf->strides[3]=0;
    fbuf->w=wd->fb.w;
    fbuf->h=wd->fb.h;
  }

  if (wd->display!=NULL && wd->fb_block!=NULL) {
    ms_mutex_unlock(&wd->xv_mutex);
    return TRUE;
  }
  ms_mutex_unlock(&wd->xv_mutex);
  return FALSE;
}

static void xv_display_lock(MSDisplay *obj){
  XvDisplay *wd = (XvDisplay*)obj->data;
  ms_mutex_lock(&wd->xv_mutex);
  if (wd->display!=NULL && wd->fb_block!=NULL)
    XLockDisplay(wd->display);
  ms_mutex_unlock(&wd->xv_mutex);
}

static void xv_display_unlock(MSDisplay *obj){
  XvDisplay *wd = (XvDisplay*)obj->data;
  ms_mutex_lock(&wd->xv_mutex);

  if (wd->display!=NULL && wd->fb_block!=NULL)
    XUnlockDisplay(wd->display);
  
  ms_mutex_unlock(&wd->xv_mutex);
}

static void xv_display_update(MSDisplay *obj, int new_image, int new_selfview){
  XvDisplay *wd = (XvDisplay*)obj->data;
  int ratiow;
  int ratioh;
  int w;
  int h;
  
  if (wd->display==NULL || wd->fb_block==NULL) {
    ms_error("no window");
    return;
  }

  ms_mutex_lock(&wd->xv_mutex);

  {
    int wposition[2];
    wposition[0]=wposition[1]=0;
    if (wd->filter) {
      ms_filter_call_method(wd->filter, MS_VIDEO_OUT_GET_WINDOW_POSITION, &wposition);
      if (wposition[0]!=wd->values_position.x || wposition[1]!=wd->values_position.y) {
        unsigned int value_mask = CWX | CWY;
        wd->values_position.x = wposition[0];
        wd->values_position.y = wposition[1];
        XConfigureWindow(wd->display, wd->window,value_mask,&wd->values_position);
      }      
    }
  }  
  
  ratiow=wd->fb.w;
  ratioh=wd->fb.h;
  ms_video_ratio_reduce(&ratiow, &ratioh);
  w = wd->screen_size.width/ratiow*ratiow;
  h = wd->screen_size.height/ratioh*ratioh;

  if (h*ratiow>w*ratioh)
  {
    w = w;
    h = w*ratioh/ratiow;
  }
  else
  {
    h = h;
    w = h*ratiow/ratioh;
  }

  if (h*wd->fb.w!=w*wd->fb.h)
    ms_error("wrong ratio");

  XLockDisplay(wd->display);
  XvPutImage( wd->display, wd->XvPort, wd->window, wd->gc,
    wd->xvImage, 0, 0, wd->xvImage->width, wd->xvImage->height, 0, 0, w, h);
  XFlush(wd->display);
  XUnlockDisplay(wd->display);
  ms_mutex_unlock(&wd->xv_mutex);
}

static int xv_poll_event(MSDisplay *obj, MSDisplayEvent *ev){
  return -1;
}

static void xv_display_uninit(MSDisplay *obj){
  XvDisplay *wd = (XvDisplay*)obj->data;
  if (wd==NULL)
    return;
  xv_show_window(wd, FALSE);
  if (wd->gc!=NULL)
    XFreeGC(wd->display, wd->gc);
  if (wd->xvImage!=NULL)
    XFree(wd->xvImage);
  if (wd->fb_block!=NULL)
    freemsg(wd->fb_block);
  if (wd->display!=NULL && wd->XvPort>0)
    XvUngrabPort(wd->display, wd->XvPort, CurrentTime);

  ms_error("destroying window!");
  if (wd->window>0 && wd->display!=NULL)
    XDestroyWindow(wd->display, wd->window);
  if (wd->display!=NULL)
    XCloseDisplay(wd->display);
  ms_free(wd);
}

MSDisplayDesc ms_xv_display_desc={
  "xv",
  .init=xv_display_init,
  .lock=xv_display_lock,
  .unlock=xv_display_unlock,
  .update=xv_display_update,
  .uninit=xv_display_uninit,
  .pollevent=xv_poll_event,
};

#endif

#if defined(HAVE_SDL13)

#include <SDL/SDL.h>
#include <SDL/SDL_video.h>

typedef struct _Sdl13Display{
  MSFilter *filter;
  bool_t sdl_initialized;
  ms_mutex_t sdl_mutex;
  SDL_Window *window;
  SDL_Renderer *renderer;
  SDL_Texture *texture;

  MSPicture fb;
  mblk_t *fb_block;	

  float sv_scalefactor;
  MSVideoSize screen_size;
} Sdl13Display;


static void sdl13_display_uninit(MSDisplay *obj);

static int sdl13_create_window(Sdl13Display *wd, int w, int h){
  static bool_t once=TRUE;
  return 0;
}

static bool_t sdl13_display_init(MSDisplay *obj, MSFilter *f, MSPicture *fbuf, MSPicture *fbuf_selfview){
  Sdl13Display *wd = (Sdl13Display*)obj->data;
  int i;
  if (wd==NULL){
    char driver[128];
    /* Initialize the SDL library */
    wd=(Sdl13Display*)ms_new0(Sdl13Display,1);
    wd->filter = f;
    obj->data=wd;

    if( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
      ms_error("Couldn't initialize SDL: %s", SDL_GetError());
      return FALSE;
    }
    SDL_LogSetAllPriority(SDL_LOG_PRIORITY_VERBOSE);
    wd->sdl_initialized=TRUE;

    if (SDL_VideoDriverName(driver, sizeof(driver))){
      ms_message("Video driver: %s", driver);
    }

    ms_mutex_init(&wd->sdl_mutex,NULL);
    ms_mutex_lock(&wd->sdl_mutex);
    wd->screen_size.width = fbuf->w;
    wd->screen_size.height = fbuf->h;

  }else {
    ms_mutex_lock(&wd->sdl_mutex);

    if (wd->texture)
      SDL_DestroyTexture(wd->texture);
    if (wd->renderer)
      SDL_DestroyRenderer(wd->renderer);
    if (wd->window)
      SDL_DestroyWindow(wd->window);
    wd->texture=NULL;
    wd->renderer=NULL;
    wd->window=NULL;

    if (wd->fb_block!=NULL)
      freemsg(wd->fb_block);
    wd->fb_block=NULL;
  }
  wd->filter = f;

  if (obj->window_id!=0)
  {
    //WindowRef window_ref = (WindowRef)obj->window_id;
    //ms_message("using windowref=%i", window_ref);
    //NSWindow *cocoa_window = [ [ NSWindow alloc ] initWithWindowRef:window_ref];
    NSView *cocoa_view = (NSView *)obj->window_id;
    NSWindow *cocoa_window = [cocoa_view window];
    if (cocoa_window==nil)
    {
      //ms_message("failure to get NSWindow (windowref=%i)", window_ref);
      wd->window = SDL_CreateWindow("Video",
        SDL_WINDOWPOS_UNDEFINED,
        SDL_WINDOWPOS_UNDEFINED,
        fbuf->w, fbuf->h,
        SDL_WINDOW_SHOWN|SDL_WINDOW_RESIZABLE);
    }
    else {
      //NSWindow *cocoa_window = (NSWindow *)obj->window_id;
      wd->window = SDL_CreateWindowFrom(cocoa_window);
    }

  }
  else
    wd->window = SDL_CreateWindow("Video",
    SDL_WINDOWPOS_UNDEFINED,
    SDL_WINDOWPOS_UNDEFINED,
    fbuf->w, fbuf->h,
    SDL_WINDOW_SHOWN|SDL_WINDOW_RESIZABLE);

  wd->renderer = SDL_CreateRenderer(wd->window, -1, 0);

  wd->texture = SDL_CreateTexture(wd->renderer,
    SDL_PIXELFORMAT_IYUV,
    SDL_TEXTUREACCESS_STREAMING,
    fbuf->w, fbuf->h);

  i=sdl13_create_window(wd, fbuf->w, fbuf->h);
  if (i==0){
#if 0
    fbuf->planes[0]=wd->lay->pixels[0];
    fbuf->planes[1]=wd->lay->pixels[2];
    fbuf->planes[2]=wd->lay->pixels[1];
    fbuf->planes[3]=NULL;
    fbuf->strides[0]=wd->lay->pitches[0];
    fbuf->strides[1]=wd->lay->pitches[2];
    fbuf->strides[2]=wd->lay->pitches[1];
    fbuf->strides[3]=0;
    fbuf->w=wd->lay->w;
    fbuf->h=wd->lay->h;
#else
    wd->fb_block = yuv_buf_alloc(&wd->fb, fbuf->w, fbuf->h);

    fbuf->planes[0]=wd->fb.planes[0];
    fbuf->planes[1]=wd->fb.planes[2];
    fbuf->planes[2]=wd->fb.planes[1];
    fbuf->planes[3]=NULL;
    fbuf->strides[0]=wd->fb.strides[0];
    fbuf->strides[1]=wd->fb.strides[2];
    fbuf->strides[2]=wd->fb.strides[1];
    fbuf->strides[3]=0;
    fbuf->w=wd->fb.w;
    fbuf->h=wd->fb.h;
#endif
    //sdl_show_window(TRUE);
    //obj->window_id=sdl_get_native_window_id();
    ms_mutex_unlock(&wd->sdl_mutex);
    return TRUE;
  }
  ms_mutex_unlock(&wd->sdl_mutex);
  return FALSE;
}

static void sdl13_display_lock(MSDisplay *obj){
  Sdl13Display *wd = (Sdl13Display*)obj->data;

  if (wd==NULL)
    return;
  ms_mutex_lock(&wd->sdl_mutex);
  //SDL_LockYUVOverlay(wd->lay);

  {
    uint8_t *surfptr;
    int32_t pitch;
    int bpp;
    uint32_t rmask, gmask, bmask, amask;

    uint32_t format;
    int access, w, h;
    SDL_QueryTexture(wd->texture, &format, &access, &w, &h);
    SDL_PixelFormatEnumToMasks(format, &bpp, &rmask, &gmask, &bmask, &amask);
    bpp = bpp / 8; /* convert to bytes per pixels */

    SDL_LockTexture(wd->texture, NULL, (void **) &surfptr, &pitch);

    ms_message("size of surface: %i", msgdsize(wd->fb_block));
    memcpy(surfptr, wd->fb_block->b_rptr, msgdsize(wd->fb_block));
  }		


  ms_mutex_unlock(&wd->sdl_mutex);
}

static void sdl13_display_unlock(MSDisplay *obj){
  Sdl13Display *wd = (Sdl13Display*)obj->data;
  if (wd==NULL)
    return;
  ms_mutex_lock(&wd->sdl_mutex);


  SDL_UnlockTexture(wd->texture);

  ms_mutex_unlock(&wd->sdl_mutex);
}

static void sdl13_display_update(MSDisplay *obj, int new_image, int new_selfview){
  Sdl13Display *wd = (Sdl13Display*)obj->data;
  SDL_Rect rect;
  int ratiow;
  int ratioh;
  int w;
  int h;

  if (wd==NULL)
    return;

  rect.x=0;
  rect.y=0;
  ms_mutex_lock(&wd->sdl_mutex);

  ratiow=wd->fb.w;
  ratioh=wd->fb.h;
  ms_video_ratio_reduce(&ratiow, &ratioh);
  w = wd->screen_size.width/ratiow*ratiow;
  h = wd->screen_size.height/ratioh*ratioh;

  if (h*ratiow>w*ratioh)
  {
    w = w;
    h = w*ratioh/ratiow;
  }
  else
  {
    h = h;
    w = h*ratiow/ratioh;
  }

  if (h*wd->fb.w!=w*wd->fb.h)
    ms_error("wrong ratio");

  rect.x = (wd->screen_size.width-w)/2;
  rect.y = (wd->screen_size.height-h)/2;
  rect.w = w;
  rect.h = h;

  //SDL_DisplayYUVOverlay(wd->lay,&rect);

  SDL_RenderCopy(wd->renderer,wd->texture, NULL, &rect);
  SDL_RenderPresent(wd->renderer);

  ms_mutex_unlock(&wd->sdl_mutex);
}

static int sdl13_poll_event(MSDisplay *obj, MSDisplayEvent *ev){
  Sdl13Display *wd = (Sdl13Display*)obj->data;
  SDL_Event event;

  if (wd==NULL) return -1;
  if (wd->window==NULL) return -1;
  ms_mutex_lock(&wd->sdl_mutex);
  if (SDL_PollEvent(&event)){
    ms_mutex_unlock(&wd->sdl_mutex);
    switch(event.type){
    case SDL_VIDEORESIZE:
      ev->evtype=MS_DISPLAY_RESIZE_EVENT;
      ev->w=event.resize.w;
      ev->h=event.resize.h;
      wd->screen_size.width = event.resize.w;
      wd->screen_size.height = event.resize.h;
      return 1;
      break;
    default:
      return 0;
      break;
    }
  }else ms_mutex_unlock(&wd->sdl_mutex);

  return -1;
}

static void sdl13_display_uninit(MSDisplay *obj){
  Sdl13Display *wd = (Sdl13Display*)obj->data;
  SDL_Event event;
  int i;
  if (wd==NULL)
    return;

  if (wd->texture)
    SDL_DestroyTexture(wd->texture);
  if (wd->renderer)
    SDL_DestroyRenderer(wd->renderer);
  if (wd->window)
    SDL_DestroyWindow(wd->window);
  wd->texture=NULL;
  wd->renderer=NULL;
  wd->window=NULL;

  if (wd->fb_block!=NULL)
    freemsg(wd->fb_block);
  wd->fb_block=NULL;
  ms_free(wd);
#ifdef __linux
  /*purge the event queue before leaving*/
  for(i=0;SDL_PollEvent(&event) && i<100;++i){
  }
#endif
  //sdl_show_window(FALSE);
  SDL_Quit();
}

MSDisplayDesc ms_sdl13_display_desc={
  "sdl13",
  .init=sdl13_display_init,
  .lock=sdl13_display_lock,
  .unlock=sdl13_display_unlock,
  .update=sdl13_display_update,
  .uninit=sdl13_display_uninit,
  .pollevent=sdl13_poll_event,
};

#endif

#if defined(HAVE_SDL)

#include <SDL/SDL.h>
#include <SDL/SDL_video.h>

typedef struct _SdlDisplay{
  MSFilter *filter;
  bool_t sdl_initialized;
  ms_mutex_t sdl_mutex;
  SDL_Surface *sdl_screen;
  SDL_Overlay *lay;

  float sv_scalefactor;
  MSVideoSize screen_size;
} SdlDisplay;

#ifdef HAVE_X11_XLIB_H

#include <SDL/SDL_syswm.h>

static long sdl_get_native_window_id(){
  SDL_SysWMinfo info;
  SDL_VERSION(&info.version);
  if ( SDL_GetWMInfo(&info) ) {
    if ( info.subsystem == SDL_SYSWM_X11 ) {
      return (long) info.info.x11.wmwindow;
    }
  }
  return 0;
}

static void sdl_show_window(bool_t show){
  SDL_SysWMinfo info;
  SDL_VERSION(&info.version);
  if ( SDL_GetWMInfo(&info) ) {
    if ( info.subsystem == SDL_SYSWM_X11 ) {
      Display *display;
      Window window;

      info.info.x11.lock_func();
      display = info.info.x11.display;
      window = info.info.x11.wmwindow;
      if (show)
        XMapWindow(display,window);
      else
        XUnmapWindow(display,window);
      info.info.x11.unlock_func();
    }
  }
}

#else

static void sdl_show_window(bool_t show){
  ms_warning("SDL window show/hide not implemented");
}

static long sdl_get_native_window_id(){
  ms_warning("sdl_get_native_window_id not implemented");
  return 0;
}

#endif

static void sdl_display_uninit(MSDisplay *obj);

static int sdl_create_window(SdlDisplay *wd, int w, int h){
  static bool_t once=TRUE;
  uint32_t flags = SDL_ANYFORMAT | SDL_DOUBLEBUF | SDL_RESIZABLE;
  const SDL_VideoInfo *info;
  info =SDL_GetVideoInfo();
  if (info->wm_available) {
    ms_message("Using window manager");
  }
  if (info->hw_available) {
    ms_message("hw surface available (%dk memory)", info->video_mem);
    flags |= SDL_HWSURFACE;
  }
  else
    flags |= SDL_SWSURFACE;

  if (info->blit_hw) {
    ms_message("hw surface available (%dk memory)", info->video_mem);
    flags |= SDL_ASYNCBLIT;
  }
  if (info->blit_hw_CC)
    ms_message("Colorkey blits between hw surfaces: accelerated");
  if (info->blit_hw_A)
    ms_message("Alpha blits between hw surfaces:  accelerated");
  if (info->blit_sw)
    ms_message("Copy blits from sw to hw surfaces:  accelerated");
  if (info->blit_hw_CC)
    ms_message("Colorkey blits between sw to hw surfaces: accelerated");
  if (info->blit_hw_A)
    ms_message("Alpha blits between sw to hw surfaces: accelerated");

  wd->sdl_screen = SDL_SetVideoMode(wd->screen_size.width,wd->screen_size.height, 0,flags);
  if (wd->sdl_screen == NULL ) {
    ms_warning("no hardware for video mode: %s\n",
      SDL_GetError());
  }
  if (wd->sdl_screen->flags & SDL_HWSURFACE) ms_message("SDL surface created in hardware");
  if (once) {
    SDL_WM_SetCaption("Video window", NULL);
    once=FALSE;
  }
  wd->lay=SDL_CreateYUVOverlay(w , h ,SDL_YV12_OVERLAY,wd->sdl_screen);
  if (wd->lay==NULL){
    ms_warning("Couldn't create yuv overlay: %s\n",
      SDL_GetError());
    return -1;
  }else{
    ms_message("%i x %i YUV overlay created: hw_accel=%i, pitches=%i,%i,%i",wd->lay->w,wd->lay->h,wd->lay->hw_overlay,
      wd->lay->pitches[0],wd->lay->pitches[1],wd->lay->pitches[2]);
    ms_message("planes= %p %p %p  %i %i",wd->lay->pixels[0],wd->lay->pixels[1],wd->lay->pixels[2],
      wd->lay->pixels[1]-wd->lay->pixels[0],wd->lay->pixels[2]-wd->lay->pixels[1]);
  }
  //SDL_ShowCursor(0);//Hide the mouse cursor if was displayed
  return 0;
}

static bool_t sdl_display_init(MSDisplay *obj, MSFilter *f, MSPicture *fbuf, MSPicture *fbuf_selfview){
  SdlDisplay *wd = (SdlDisplay*)obj->data;
  int i;
  if (wd==NULL){
    char driver[128];
    /* Initialize the SDL library */
    wd=(SdlDisplay*)ms_new0(SdlDisplay,1);
    wd->filter = f;
    obj->data=wd;

    if( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
      ms_error("Couldn't initialize SDL: %s", SDL_GetError());
      return FALSE;
    }
    wd->sdl_initialized=TRUE;
    if (SDL_VideoDriverName(driver, sizeof(driver))){
      ms_message("Video driver: %s", driver);
    }
    ms_mutex_init(&wd->sdl_mutex,NULL);
    ms_mutex_lock(&wd->sdl_mutex);
    wd->screen_size.width = fbuf->w;
    wd->screen_size.height = fbuf->h;

  }else {
    ms_mutex_lock(&wd->sdl_mutex);

    if (wd->lay!=NULL)
      SDL_FreeYUVOverlay(wd->lay);
    if (wd->sdl_screen!=NULL)
      SDL_FreeSurface(wd->sdl_screen);
    wd->lay=NULL;
    wd->sdl_screen=NULL;
  }
  wd->filter = f;

  i=sdl_create_window(wd, fbuf->w, fbuf->h);
  if (i==0){
    fbuf->planes[0]=wd->lay->pixels[0];
    fbuf->planes[1]=wd->lay->pixels[2];
    fbuf->planes[2]=wd->lay->pixels[1];
    fbuf->planes[3]=NULL;
    fbuf->strides[0]=wd->lay->pitches[0];
    fbuf->strides[1]=wd->lay->pitches[2];
    fbuf->strides[2]=wd->lay->pitches[1];
    fbuf->strides[3]=0;
    fbuf->w=wd->lay->w;
    fbuf->h=wd->lay->h;
    sdl_show_window(TRUE);
    obj->window_id=sdl_get_native_window_id();
    ms_mutex_unlock(&wd->sdl_mutex);
    return TRUE;
  }
  ms_mutex_unlock(&wd->sdl_mutex);
  return FALSE;
}

static void sdl_display_lock(MSDisplay *obj){
  SdlDisplay *wd = (SdlDisplay*)obj->data;
  if (wd==NULL)
    return;
  ms_mutex_lock(&wd->sdl_mutex);
  SDL_LockYUVOverlay(wd->lay);
  ms_mutex_unlock(&wd->sdl_mutex);
}

static void sdl_display_unlock(MSDisplay *obj){
  SdlDisplay *wd = (SdlDisplay*)obj->data;
  if (wd==NULL)
    return;
  ms_mutex_lock(&wd->sdl_mutex);
  SDL_UnlockYUVOverlay(wd->lay);
  ms_mutex_unlock(&wd->sdl_mutex);
}

static void sdl_display_update(MSDisplay *obj, int new_image, int new_selfview){
  SdlDisplay *wd = (SdlDisplay*)obj->data;
  SDL_Rect rect;
  int ratiow;
  int ratioh;
  int w;
  int h;

  if (wd==NULL)
    return;

  rect.x=0;
  rect.y=0;
  ms_mutex_lock(&wd->sdl_mutex);

  ratiow=wd->lay->w;
  ratioh=wd->lay->h;
  ms_video_ratio_reduce(&ratiow, &ratioh);
  w = wd->screen_size.width/ratiow*ratiow;
  h = wd->screen_size.height/ratioh*ratioh;

  if (h*ratiow>w*ratioh)
  {
    w = w;
    h = w*ratioh/ratiow;
  }
  else
  {
    h = h;
    w = h*ratiow/ratioh;
  }

  if (h*wd->lay->w!=w*wd->lay->h)
    ms_error("wrong ratio");

  rect.x = (wd->screen_size.width-w)/2;
  rect.y = (wd->screen_size.height-h)/2;
  rect.w = w;
  rect.h = h;

  SDL_DisplayYUVOverlay(wd->lay,&rect);
  ms_mutex_unlock(&wd->sdl_mutex);
}

static int sdl_poll_event(MSDisplay *obj, MSDisplayEvent *ev){
  SdlDisplay *wd = (SdlDisplay*)obj->data;
  SDL_Event event;
  if (wd==NULL) return -1;
  if (wd->sdl_screen==NULL) return -1;
  ms_mutex_lock(&wd->sdl_mutex);
  if (SDL_PollEvent(&event)){
    ms_mutex_unlock(&wd->sdl_mutex);
    switch(event.type){
    case SDL_VIDEORESIZE:
      ev->evtype=MS_DISPLAY_RESIZE_EVENT;
      ev->w=event.resize.w;
      ev->h=event.resize.h;
      wd->screen_size.width = event.resize.w;
      wd->screen_size.height = event.resize.h;
      return 1;
      break;
    default:
      return 0;
      break;
    }
  }else ms_mutex_unlock(&wd->sdl_mutex);
  return -1;
}

static void sdl_display_uninit(MSDisplay *obj){
  SdlDisplay *wd = (SdlDisplay*)obj->data;
  SDL_Event event;
  int i;
  if (wd==NULL)
    return;
  if (wd->lay!=NULL)
    SDL_FreeYUVOverlay(wd->lay);
  if (wd->sdl_screen!=NULL){
    SDL_FreeSurface(wd->sdl_screen);
    wd->sdl_screen=NULL;
  }
  wd->lay=NULL;
  wd->sdl_screen=NULL;
  ms_free(wd);
#ifdef __linux
  /*purge the event queue before leaving*/
  for(i=0;SDL_PollEvent(&event) && i<100;++i){
  }
#endif
  sdl_show_window(FALSE);
  SDL_Quit();
}

MSDisplayDesc ms_sdl_display_desc={
  "sdl",
  .init=sdl_display_init,
  .lock=sdl_display_lock,
  .unlock=sdl_display_unlock,
  .update=sdl_display_update,
  .uninit=sdl_display_uninit,
  .pollevent=sdl_poll_event,
};

#endif

#if defined(NOWINDOW_DISPLAY)

typedef struct NoWindowDisplay{
  MSFilter *filter;
  MSPicture fb;
  MSPicture fb_selfview;
  uint8_t *rgb_selfview;
  int rgb_len_selfview;
  struct MSScalerContext *sws_selfview;
  uint8_t *rgb;
  int rgb_len;
  struct MSScalerContext *sws;
  MSPixFmt format;
}NoWindowDisplay;


static bool_t nowindow_display_init(MSDisplay *obj, MSFilter *f, MSPicture *fbuf, MSPicture *fbuf_selfview){
  NoWindowDisplay *wd=(NoWindowDisplay*)obj->data;
  int ysize,usize;

  if (wd!=NULL)
  {
    wd->filter = NULL;
    if (wd->fb.planes[0]) ms_free(wd->fb.planes[0]);
    wd->fb.planes[0]=NULL;
    wd->fb.planes[1]=NULL;
    wd->fb.planes[2]=NULL;
    wd->fb.planes[3]=NULL;
    if (wd->rgb) ms_free(wd->rgb);
    wd->rgb=NULL;
    wd->rgb_len=0;
    if (wd->sws!=NULL) ms_video_scalercontext_free(wd->sws);
    else
    {
      ms_message("We never had any mainview images");
    }
    wd->sws=NULL;
    if (wd->fb_selfview.planes[0]) ms_free(wd->fb_selfview.planes[0]);
    wd->fb_selfview.planes[0]=NULL;
    wd->fb_selfview.planes[1]=NULL;
    wd->fb_selfview.planes[2]=NULL;
    wd->fb_selfview.planes[3]=NULL;
    if (wd->rgb_selfview) ms_free(wd->rgb_selfview);
    wd->rgb_selfview=NULL;
    wd->rgb_len_selfview=0;
    if (wd->sws_selfview!=NULL) ms_video_scalercontext_free(wd->sws_selfview);
    else
    {
      ms_message("We never had any selfview images");
    }
    wd->sws_selfview=NULL;
  }
  else
    wd=(NoWindowDisplay*)ms_new0(NoWindowDisplay,1);

#if defined(NOWINDOW_DISPLAY_ONEVIEW)
  wd->format = MS_RGBA;
#elif defined(NOWINDOW_DISPLAY_UYVY)
  wd->format = MS_UYVY;
#elif defined(NOWINDOW_DISPLAY_YUV420P)
  wd->format = MS_YUV420P;
#elif defined(WIN32)
  wd->format = MS_RGB24;
#else
  wd->format = MS_RGBA;
#endif

  wd->filter = f;
  obj->data=wd;

  wd->fb.w=fbuf->w;
  wd->fb.h=fbuf->h;
  wd->fb_selfview.w=fbuf_selfview->w;
  wd->fb_selfview.h=fbuf_selfview->h;

  /*allocate yuv and rgb buffers*/
  if (wd->fb_selfview.planes[0]) ms_free(wd->fb_selfview.planes[0]);
  if (wd->rgb_selfview) ms_free(wd->rgb_selfview);
  ysize=wd->fb_selfview.w*wd->fb_selfview.h;
  usize=ysize/4;
  fbuf_selfview->planes[0]=wd->fb_selfview.planes[0]=(uint8_t*)ms_malloc0(ysize+2*usize);
  fbuf_selfview->planes[1]=wd->fb_selfview.planes[1]=wd->fb_selfview.planes[0]+ysize;
  fbuf_selfview->planes[2]=wd->fb_selfview.planes[2]=wd->fb_selfview.planes[1]+usize;
  fbuf_selfview->planes[3]=NULL;
  fbuf_selfview->strides[0]=wd->fb_selfview.strides[0]=wd->fb_selfview.w;
  fbuf_selfview->strides[1]=wd->fb_selfview.strides[1]=wd->fb_selfview.w/2;
  fbuf_selfview->strides[2]=wd->fb_selfview.strides[2]=wd->fb_selfview.w/2;
  fbuf_selfview->strides[3]=0;

  if (wd->format == MS_YUV420P) {
    wd->rgb_len=0;
    wd->rgb=NULL;
  } else if (wd->format == MS_UYVY) {
    wd->rgb_len_selfview=ysize*2;
    wd->rgb_selfview=(uint8_t*)ms_malloc0(wd->rgb_len_selfview);
  } else if (wd->format == MS_RGBA) {
    wd->rgb_len_selfview=ysize*4;
    wd->rgb_selfview=(uint8_t*)ms_malloc0(wd->rgb_len_selfview);
  } else { /* MS_RGB24 */
#ifdef WIN32
    /* rgb_stride must be multiple of 4... */
    int rgb_stride=wd->fb_selfview.w*3;
    rgb_stride = (rgb_stride + 3) & ~0x3;
    wd->rgb_len_selfview=rgb_stride*wd->fb_selfview.h;
#else
    wd->rgb_len_selfview=ysize*3;
#endif

    wd->rgb_selfview=(uint8_t*)ms_malloc0(wd->rgb_len_selfview);
  }

  if (wd->fb.planes[0]) ms_free(wd->fb.planes[0]);
  if (wd->rgb) ms_free(wd->rgb);
  ysize=wd->fb.w*wd->fb.h;
  usize=ysize/4;
  fbuf->planes[0]=wd->fb.planes[0]=(uint8_t*)ms_malloc0(ysize+2*usize);
  fbuf->planes[1]=wd->fb.planes[1]=wd->fb.planes[0]+ysize;
  fbuf->planes[2]=wd->fb.planes[2]=wd->fb.planes[1]+usize;
  fbuf->planes[3]=NULL;
  fbuf->strides[0]=wd->fb.strides[0]=wd->fb.w;
  fbuf->strides[1]=wd->fb.strides[1]=wd->fb.w/2;
  fbuf->strides[2]=wd->fb.strides[2]=wd->fb.w/2;
  fbuf->strides[3]=0;

  if (wd->format == MS_YUV420P) {
    wd->rgb_len=0;
    wd->rgb=NULL;
  } else if (wd->format == MS_UYVY) {
    wd->rgb_len=ysize*2;
    wd->rgb=(uint8_t*)ms_malloc0(wd->rgb_len);
  } else if (wd->format == MS_RGBA) {
    wd->rgb_len=ysize*4;
    wd->rgb=(uint8_t*)ms_malloc0(wd->rgb_len);
  } else { /* MS_RGB24 */
#ifdef WIN32
    /* rgb_stride must be multiple of 4... */
    int rgb_stride=wd->fb.w*3;
    rgb_stride = (rgb_stride + 3) & ~0x3;
    wd->rgb_len=rgb_stride*wd->fb.h;
#else
    wd->rgb_len=ysize*3;
#endif
    wd->rgb=(uint8_t*)ms_malloc0(wd->rgb_len);
  }
  return TRUE;
}

static void yuv420p_to_rgb(NoWindowDisplay *wd, MSPicture *src, uint8_t *rgb){
  if (wd->format == MS_YUV420P) {
    return;
  } else if (wd->format == MS_UYVY)
  {
    int rgb_stride;
    uint8_t *p;

    rgb_stride=src->w*2;
    p=rgb;

    if (wd->sws==NULL){
      wd->sws=ms_video_scalercontext_init(src->w,src->h,MS_YUV420P,
        src->w,src->h, MS_UYVY,
        MS_YUVFAST, NULL, NULL, NULL);
    }
    if (ms_video_scalercontext_convert(wd->sws,src->planes,src->strides, 0,
      src->h, &p, &rgb_stride)<0){
        ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
    }
  } else if (wd->format == MS_RGBA)
  {
    int rgb_stride;
    uint8_t *p;

#ifdef WIN32
    rgb_stride=-src->w*4;
    p=rgb+(src->w*4*(src->h-1));
#else
    rgb_stride=src->w*4;
    p=rgb;
#endif

    if (wd->sws==NULL){
      wd->sws=ms_video_scalercontext_init(src->w,src->h,MS_YUV420P,
        src->w,src->h, MS_ARGB,
        MS_YUVFAST, NULL, NULL, NULL);
    }
    if (ms_video_scalercontext_convert(wd->sws,src->planes,src->strides, 0,
      src->h, &p, &rgb_stride)<0){
        ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
    }
  } else { /* MS_RGB24 */
    int rgb_stride;
    uint8_t *p;

#ifdef WIN32
    rgb_stride=src->w*3;
    /* rgb_stride must be multiple of 4... */
    rgb_stride = (rgb_stride + 3) & ~0x3;
    p=rgb;
    //old way with MS_RGB24_REV
    //p=rgb+(rgb_stride*(src->h-1));
    //rgb_stride = -rgb_stride;
#else
    rgb_stride=src->w*3;
    p=rgb;
#endif

    if (wd->sws==NULL){
      wd->sws=ms_video_scalercontext_init(src->w,src->h,MS_YUV420P,
        src->w,src->h, MS_RGB24,
        MS_YUVFAST, NULL, NULL, NULL);
    }
    if (ms_video_scalercontext_convert(wd->sws,src->planes,src->strides, 0,
      src->h, &p, &rgb_stride)<0){
        ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
    }
  }
}

static void yuv420p_to_rgb_selfview(NoWindowDisplay *wd, MSPicture *src, uint8_t *rgb){
  if (wd->format == MS_YUV420P) {
    return;
  } else if (wd->format == MS_UYVY)
  {
    int rgb_stride;
    uint8_t *p;

    rgb_stride=src->w*2;
    p=rgb;

    if (wd->sws_selfview==NULL){
      wd->sws_selfview=ms_video_scalercontext_init(src->w,src->h,MS_YUV420P,
        src->w,src->h, MS_UYVY,
        MS_YUVFAST, NULL, NULL, NULL);
    }
    if (ms_video_scalercontext_convert(wd->sws_selfview,src->planes,src->strides, 0,
      src->h, &p, &rgb_stride)<0){
        ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
    }
  } else if (wd->format == MS_RGBA)
  {
    int rgb_stride;
    uint8_t *p;

#ifdef WIN32
    rgb_stride=-src->w*4;
    p=rgb+(src->w*4*(src->h-1));
#else
    rgb_stride=src->w*4;
    p=rgb;
#endif

    if (wd->sws_selfview==NULL){
      wd->sws_selfview=ms_video_scalercontext_init(src->w,src->h,MS_YUV420P,
        src->w,src->h, MS_ARGB,
        MS_YUVFAST, NULL, NULL, NULL);
    }
    if (ms_video_scalercontext_convert(wd->sws_selfview,src->planes,src->strides, 0,
      src->h, &p, &rgb_stride)<0){
        ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
    }
  } else { /* MS_RGB24 */
    int rgb_stride;
    uint8_t *p;

#ifdef WIN32
    rgb_stride=src->w*3;
    /* rgb_stride must be multiple of 4... */
    rgb_stride = (rgb_stride + 3) & ~0x3;
    p = rgb;
    //old way with MS_RGB24_REV
    //p=rgb+(rgb_stride*(src->h-1));
    //rgb_stride = -rgb_stride;
#else
    rgb_stride=src->w*3;
    p=rgb;
#endif

    if (wd->sws_selfview==NULL){
      wd->sws_selfview=ms_video_scalercontext_init(src->w,src->h,MS_YUV420P,
        src->w,src->h, MS_RGB24,
        MS_YUVFAST, NULL, NULL, NULL);
    }
    if (ms_video_scalercontext_convert(wd->sws_selfview,src->planes,src->strides, 0,
      src->h, &p, &rgb_stride)<0){
        ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
    }
  }
}

static void nowindow_display_update(MSDisplay *obj, int new_image, int new_selfview){
  struct amsip_picture {
    int width;
    int height;
    int format;
    int size;
    unsigned char *pixels;
  };
  NoWindowDisplay *wd=(NoWindowDisplay*)obj->data;

  if (wd->format == MS_YUV420P) {
    if (wd->filter==NULL)
      return;
    if (new_image>0)
    {        
      struct amsip_picture pic;
      pic.width = wd->fb.w;
      pic.height = wd->fb.h;
      pic.format = wd->format;
      pic.size = wd->fb.w*wd->fb.h*3/2;
      pic.pixels = wd->fb.planes[0];
      ms_filter_notify(wd->filter, MS_DISPLAY_NEW_IMAGE, (void*)&pic);
    }
    if (new_selfview>0)
    {
      struct amsip_picture pic;
      pic.width = wd->fb_selfview.w;
      pic.height = wd->fb_selfview.h;
      pic.format = wd->format;
      pic.size = wd->fb_selfview.w*wd->fb_selfview.h*3/2;
      pic.pixels = wd->fb_selfview.planes[0];
      ms_filter_notify(wd->filter, MS_DISPLAY_NEW_IMAGE_SELFVIEW, (void*)&pic);
    }
    return;
  }

  if (wd->format == MS_UYVY)
  {
    if (wd->filter==NULL)
      return;
    if (new_image>0)
    {        
      struct amsip_picture pic;
      yuv420p_to_rgb(wd, &wd->fb, wd->rgb);
      pic.width = wd->fb.w;
      pic.height = wd->fb.h;
      pic.format = wd->format;
      pic.size = wd->rgb_len;
      pic.pixels = wd->rgb;
      ms_filter_notify(wd->filter, MS_DISPLAY_NEW_IMAGE, (void*)&pic);
    }
    if (new_selfview>0)
    {
      struct amsip_picture pic;
      yuv420p_to_rgb_selfview(wd, &wd->fb_selfview, wd->rgb_selfview);

      pic.width = wd->fb_selfview.w;
      pic.height = wd->fb_selfview.h;
      pic.format = wd->format;
      pic.size = wd->rgb_len_selfview;
      pic.pixels = wd->rgb_selfview;
      ms_filter_notify(wd->filter, MS_DISPLAY_NEW_IMAGE_SELFVIEW, (void*)&pic);
    }
    return;
  }

  if (new_image>0)
  {
    if (wd->filter!=NULL)
    {
#if (!defined(NOWINDOW_DISPLAY_ONEVIEW)) ||  (TARGET_OS_IPHONE==1) || (TARGET_IPHONE_SIMULATOR==1)
      struct amsip_picture pic;
#endif
      yuv420p_to_rgb(wd, &wd->fb, wd->rgb);

#if (!defined(NOWINDOW_DISPLAY_ONEVIEW)) ||  (TARGET_OS_IPHONE==1) || (TARGET_IPHONE_SIMULATOR==1)
      pic.width = wd->fb.w;
      pic.height = wd->fb.h;
      pic.format = wd->format;
      pic.size = wd->rgb_len;
      pic.pixels = wd->rgb;
      ms_filter_notify(wd->filter, MS_DISPLAY_NEW_IMAGE, (void*)&pic);
#endif
    }
  }

#if defined(NOWINDOW_DISPLAY_ONEVIEW) && (TARGET_OS_IPHONE==0) && (TARGET_IPHONE_SIMULATOR==0)
  if (new_selfview>0 && new_image<=0)
    return;

  if (new_image>0)
  {
    struct amsip_picture pic;
    int corner=0;

    if (wd->filter)
      ms_filter_call_method(wd->filter, MS_VIDEO_OUT_GET_CORNER, &corner);

    if (corner>=0)
    {
      float sv_scalefactor=SCALE_FACTOR;
      int w_selfview;
      int h_selfview;
      int ratiow;
      int ratioh;
      int rgb_stride;
      uint8_t *p;

      if (wd->filter)
        ms_filter_call_method(wd->filter, MS_VIDEO_OUT_GET_SCALE_FACTOR, &sv_scalefactor);
      /* preserve ratio */
      ratiow=wd->fb_selfview.w;
      ratioh=wd->fb_selfview.h;
      ms_video_ratio_reduce(&ratiow, &ratioh);

      w_selfview = (int)(wd->fb_selfview.w/sv_scalefactor);
      w_selfview = w_selfview/ratiow*ratiow;
      h_selfview = w_selfview*ratioh/ratiow;

      if (wd->sws_selfview==NULL){
        wd->sws_selfview=ms_video_scalercontext_init(wd->fb_selfview.w,wd->fb_selfview.h,MS_YUV420P,
          w_selfview,h_selfview, MS_ARGB,
          MS_YUVFAST, NULL, NULL, NULL);
      }


      if (corner==0 || corner==4)
      {
        /*bottom right */
        rgb_stride=wd->fb.w*4;
        p=wd->rgb+(rgb_stride)*wd->fb.h-(rgb_stride)*h_selfview+(rgb_stride)-w_selfview*4;
      } else if (corner==2 || corner==6)
      {
        /*top right */
        rgb_stride=wd->fb.w*4;
        p=wd->rgb+(rgb_stride)-w_selfview*4;
      } else if (corner==3 || corner==7)
      {
        /* bottom left: OK */
        rgb_stride=wd->fb.w*4;
        p=wd->rgb+(rgb_stride)*wd->fb.h-(rgb_stride)*h_selfview;
      } else {
        /* top left: OK */
        rgb_stride=wd->fb_selfview.w*4;
        p=wd->rgb;
      }

      if (ms_video_scalercontext_convert(wd->sws_selfview,wd->fb_selfview.planes,wd->fb_selfview.strides, 0,
        wd->fb_selfview.h, &p, &rgb_stride)<0){
          ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
      }
    }

    pic.width = wd->fb.w;
    pic.height = wd->fb.h;
    pic.format = wd->format;
    pic.size = wd->rgb_len;
    pic.pixels = wd->rgb;
    ms_filter_notify(wd->filter, MS_DISPLAY_NEW_IMAGE, (void*)&pic);		
  }
#else
  if (new_selfview>0)
  {
    if (wd->filter!=NULL)
    {
      struct amsip_picture pic;
      yuv420p_to_rgb_selfview(wd, &wd->fb_selfview, wd->rgb_selfview);

      pic.width = wd->fb.w;
      pic.height = wd->fb.h;
      pic.format = wd->format;
      pic.size = wd->rgb_len_selfview;
      pic.pixels = wd->rgb_selfview;
      ms_filter_notify(wd->filter, MS_DISPLAY_NEW_IMAGE_SELFVIEW, (void*)&pic);
    }
  }
#endif
}

static void nowindow_display_uninit(MSDisplay *obj){
  NoWindowDisplay *wd=(NoWindowDisplay*)obj->data;
  if (wd==NULL)
    return;
  if (wd->fb_selfview.planes[0]) ms_free(wd->fb_selfview.planes[0]);
  if (wd->rgb_selfview) ms_free(wd->rgb_selfview);
  if (wd->sws_selfview) ms_video_scalercontext_free(wd->sws_selfview);
  if (wd->fb.planes[0]) ms_free(wd->fb.planes[0]);
  if (wd->rgb) ms_free(wd->rgb);
  if (wd->sws) ms_video_scalercontext_free(wd->sws);
  ms_free(wd);
}

static int nowindow_display_pollevent(MSDisplay *d, MSDisplayEvent *ev){
  return -1;
}

MSDisplayDesc ms_nowindow_display_desc={
  "callback",
  nowindow_display_init,
  NULL,
  NULL,
  nowindow_display_update,
  nowindow_display_uninit,
  nowindow_display_pollevent
};

#endif

#if defined(WIN32)

#include <Vfw.h>

typedef struct _WinDisplay{
  MSFilter *filter;
  HWND window;
  HDRAWDIB ddh;
  MSPicture fb;
  MSPicture fb_selfview;
  uint8_t *rgb_selfview;
  int rgb_len_selfview;
  struct MSScalerContext *sws_selfview;
  MSDisplayEvent last_rsz;
  uint8_t *rgb;
  int last_rect_w;
  int last_rect_h;
  int rgb_len;
  struct MSScalerContext *sws;
  bool_t new_ev;
}WinDisplay;

static LRESULT CALLBACK window_proc(
  HWND hwnd,        // handle to window
  UINT uMsg,        // message identifier
  WPARAM wParam,    // first message parameter
  LPARAM lParam)    // second message parameter
{
  switch(uMsg){
  case WM_DESTROY:
    break;
  case WM_SIZE:
    if (wParam==SIZE_RESTORED){
      int h=(lParam>>16) & 0xffff;
      int w=lParam & 0xffff;
      MSDisplay *obj;
      WinDisplay *wd;
      ms_message("Resized to %i,%i",w,h);
      obj=(MSDisplay*)GetWindowLongPtr(hwnd,GWLP_USERDATA);
      if (obj!=NULL){
        wd=(WinDisplay*)obj->data;
        wd->last_rsz.evtype=MS_DISPLAY_RESIZE_EVENT;
        wd->last_rsz.w=w;
        wd->last_rsz.h=h;
        wd->new_ev=TRUE;
      }else{
        ms_error("Could not retrieve MSDisplay from window !");
      }
    }
    break;
#if 0
  case WM_LBUTTONDOWN:
    ms_message("WM_LBUTTONDOWN");
    break;
  case WM_LBUTTONUP:
    ms_message("WM_LBUTTONUP");
    break;
  case WM_LBUTTONDBLCLK:
    ms_message("WM_LBUTTONDBLCLK");
    break;
  case WM_RBUTTONDOWN:
    ms_message("WM_RBUTTONDOWN");
    break;
  case WM_RBUTTONUP:
    ms_message("WM_RBUTTONUP");
    break;
  case WM_RBUTTONDBLCLK:
    ms_message("WM_RBUTTONDBLCLK");
    break;
  case WM_MOUSEMOVE:
    ms_message("WM_MOUSEMOVE");
    break;
  case WM_MOUSEWHEEL:
    ms_message("WM_MOUSEWHEEL");
    break;
  case WM_KEYDOWN:
    ms_message("WM_KEYDOWN");
    break;
  case WM_KEYUP:
    ms_message("WM_KEYUP");
    break;
  case 0X102:
    ms_message("0X102");
    break;
  case 0x20:
    ms_message("0x20");
    break;
#endif  
  default:
    return DefWindowProc(hwnd, uMsg, wParam, lParam);
  }
  return 0;
}

static HWND create_window(int w, int h)
{
  WNDCLASS wc;
  HINSTANCE hInstance = GetModuleHandle(NULL);
  HWND hwnd;
  RECT rect;
  wc.style = 0 ;
  wc.lpfnWndProc = window_proc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = NULL;
  wc.hIcon = NULL;
  wc.hCursor = LoadCursor(hInstance, IDC_ARROW);
  wc.hbrBackground = NULL;
  wc.lpszMenuName =  NULL;
  wc.lpszClassName = L"Video Window";

  if(!RegisterClass(&wc))
  {
    /* already registred! */
  }
  rect.left=100;
  rect.top=100;
  rect.right=rect.left+w;
  rect.bottom=rect.top+h;
  if (!AdjustWindowRect(&rect,WS_OVERLAPPEDWINDOW|WS_VISIBLE /*WS_CAPTION WS_TILED|WS_BORDER*/,FALSE)){
    ms_error("AdjustWindowRect failed.");
  }
  
  ms_message("AdjustWindowRect: %li,%li %li,%li",rect.left,rect.top,rect.right,rect.bottom);
  hwnd=CreateWindow(L"Video Window", L"Video window", 
    WS_OVERLAPPEDWINDOW /*WS_THICKFRAME*/ | WS_VISIBLE ,
    CW_USEDEFAULT, CW_USEDEFAULT, rect.right-rect.left,rect.bottom-rect.top,
    NULL, NULL, hInstance, NULL);
  if (hwnd==NULL){
    ms_error("Fail to create video window");
  }
  return hwnd;
}

#define USE_RGB32BIT
static void yuv420p_to_rgb(WinDisplay *wd, MSPicture *src, uint8_t *rgb){
#ifdef USE_RGB32BIT
  int rgb_stride=-src->w*4;
  uint8_t *p;

  p=rgb+(src->w*4*(src->h-1));
  if (wd->sws==NULL){
    wd->sws=ms_video_scalercontext_init(src->w,src->h,MS_YUV420P,
      src->w,src->h, MS_BGRA,
      MS_YUVFAST, NULL, NULL, NULL);
  }
  if (ms_video_scalercontext_convert(wd->sws,src->planes,src->strides, 0,
    src->h, &p, &rgb_stride)<0){
      ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
  }
#else
  int rgb_stride=-src->w*3;
  uint8_t *p;

  p=rgb+(src->w*3*(src->h-1));
  if (wd->sws==NULL){
    wd->sws=ms_video_scalercontext_init(src->w,src->h,MS_YUV420P,
      src->w,src->h, MS_RGB24_REV,
      MS_YUVFAST, NULL, NULL, NULL);
  }
  if (ms_video_scalercontext_convert(wd->sws,src->planes,src->strides, 0,
    src->h, &p, &rgb_stride)<0){
      ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
  }
#endif
}

static void yuv420p_to_rgb_selfview(WinDisplay *wd, MSPicture *src, uint8_t *rgb){
#ifdef USE_RGB32BIT
  int rgb_stride=-src->w*4;
  uint8_t *p;

  p=rgb+(src->w*4*(src->h-1));
  if (wd->sws_selfview==NULL){
    wd->sws_selfview=ms_video_scalercontext_init(src->w,src->h,MS_YUV420P,
      src->w,src->h, MS_BGRA,
      MS_YUVFAST, NULL, NULL, NULL);
  }
  if (ms_video_scalercontext_convert(wd->sws_selfview,src->planes,src->strides, 0,
    src->h, &p, &rgb_stride)<0){
      ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
  }
#else
  int rgb_stride=-src->w*3;
  uint8_t *p;

  p=rgb+(src->w*3*(src->h-1));
  if (wd->sws_selfview==NULL){
    wd->sws_selfview=ms_video_scalercontext_init(src->w,src->h,MS_YUV420P,
      src->w,src->h, MS_RGB24_REV,
      MS_YUVFAST, NULL, NULL, NULL);
  }
  if (ms_video_scalercontext_convert(wd->sws_selfview,src->planes,src->strides, 0,
    src->h, &p, &rgb_stride)<0){
      ms_error("Error in 420->rgb ms_video_scalercontext_convert().");
  }
#endif
}

static bool_t win_display_init(MSDisplay *obj, MSFilter *f, MSPicture *fbuf, MSPicture *fbuf_selfview){
  WinDisplay *wd=(WinDisplay*)obj->data;
  int ysize,usize;

  if (wd!=NULL)
  {
    wd->filter = NULL;
    if (wd->ddh) DrawDibClose(wd->ddh);
    wd->ddh=NULL;
    if (wd->fb.planes[0]) ms_free(wd->fb.planes[0]);
    wd->fb.planes[0]=NULL;
    wd->fb.planes[1]=NULL;
    wd->fb.planes[2]=NULL;
    wd->fb.planes[3]=NULL;
    if (wd->rgb) ms_free(wd->rgb);
    wd->rgb=NULL;
    wd->rgb_len=0;
    if (wd->sws!=NULL) ms_video_scalercontext_free(wd->sws);
    else
    {
      ms_message("We never had any mainview images");
    }
    wd->sws=NULL;
    if (wd->fb_selfview.planes[0]) ms_free(wd->fb_selfview.planes[0]);
    wd->fb_selfview.planes[0]=NULL;
    wd->fb_selfview.planes[1]=NULL;
    wd->fb_selfview.planes[2]=NULL;
    wd->fb_selfview.planes[3]=NULL;
    if (wd->rgb_selfview) ms_free(wd->rgb_selfview);
    wd->rgb_selfview=NULL;
    wd->rgb_len_selfview=0;
    if (wd->sws_selfview!=NULL) ms_video_scalercontext_free(wd->sws_selfview);
    else
    {
      ms_message("We never had any selfview images");
    }
    wd->sws_selfview=NULL;
    wd->last_rect_w=0;
    wd->last_rect_h=0;
  }
  else
    wd=(WinDisplay*)ms_new0(WinDisplay,1);

  wd->filter = f;
  obj->data=wd;

  wd->fb.w=fbuf->w;
  wd->fb.h=fbuf->h;
  wd->fb_selfview.w=fbuf_selfview->w;
  wd->fb_selfview.h=fbuf_selfview->h;

  if (wd->window==NULL){
    if (obj->use_external_window && obj->window_id!=0){
      void *p;
      wd->window=(HWND)obj->window_id;
      p=(void*)GetWindowLongPtr(wd->window,GWLP_USERDATA);
      if (p!=NULL){
        ms_error("Gulp: this externally supplied windows seems to "
          "already have a userdata ! resizing will crash !");
      }else SetWindowLongPtr(wd->window,GWLP_USERDATA,(LONG_PTR)obj);
    }else{
      wd->window=create_window(wd->fb.w,wd->fb.h);
      obj->window_id=(long)wd->window;
      if (wd->window!=NULL) SetWindowLongPtr(wd->window,GWLP_USERDATA,(LONG_PTR)obj);
      else return FALSE;
    }
  }else if (!obj->use_external_window){
    /* the window might need to be resized*/
    RECT cur;
    GetWindowRect(wd->window,&cur);
    MoveWindow(wd->window,cur.left, cur.top, wd->fb.w, wd->fb.h,TRUE);
  }

  if (wd->ddh==NULL) wd->ddh=DrawDibOpen();
  if (wd->ddh==NULL){
    ms_error("DrawDibOpen() failed.");
    return FALSE;
  }

  /*allocate yuv and rgb buffers*/
  if (wd->fb_selfview.planes[0]) ms_free(wd->fb_selfview.planes[0]);
  if (wd->rgb_selfview) ms_free(wd->rgb_selfview);
  ysize=wd->fb_selfview.w*wd->fb_selfview.h;
  usize=ysize/4;
  fbuf_selfview->planes[0]=wd->fb_selfview.planes[0]=(uint8_t*)ms_malloc0(ysize+2*usize);
  fbuf_selfview->planes[1]=wd->fb_selfview.planes[1]=wd->fb_selfview.planes[0]+ysize;
  fbuf_selfview->planes[2]=wd->fb_selfview.planes[2]=wd->fb_selfview.planes[1]+usize;
  fbuf_selfview->planes[3]=NULL;
  fbuf_selfview->strides[0]=wd->fb_selfview.strides[0]=wd->fb_selfview.w;
  fbuf_selfview->strides[1]=wd->fb_selfview.strides[1]=wd->fb_selfview.w/2;
  fbuf_selfview->strides[2]=wd->fb_selfview.strides[2]=wd->fb_selfview.w/2;
  fbuf_selfview->strides[3]=0;

#ifdef USE_RGB32BIT
  wd->rgb_len_selfview=ysize*4;
#else
  wd->rgb_len_selfview=ysize*3;
#endif
  wd->rgb_selfview=(uint8_t*)ms_malloc0(wd->rgb_len_selfview);

  if (wd->fb.planes[0]) ms_free(wd->fb.planes[0]);
  if (wd->rgb) ms_free(wd->rgb);
  ysize=wd->fb.w*wd->fb.h;
  usize=ysize/4;
  fbuf->w=wd->fb.w;
  fbuf->h=wd->fb.h;
  fbuf->planes[0]=wd->fb.planes[0]=(uint8_t*)ms_malloc0(ysize+2*usize);
  fbuf->planes[1]=wd->fb.planes[1]=wd->fb.planes[0]+ysize;
  fbuf->planes[2]=wd->fb.planes[2]=wd->fb.planes[1]+usize;
  fbuf->planes[3]=NULL;
  fbuf->strides[0]=wd->fb.strides[0]=wd->fb.w;
  fbuf->strides[1]=wd->fb.strides[1]=wd->fb.w/2;
  fbuf->strides[2]=wd->fb.strides[2]=wd->fb.w/2;
  fbuf->strides[3]=0;

#ifdef USE_RGB32BIT
  wd->rgb_len=ysize*4;
#else
  wd->rgb_len=ysize*3;
#endif
  wd->rgb=(uint8_t*)ms_malloc0(wd->rgb_len);
  wd->last_rect_w=0;
  wd->last_rect_h=0;

  {
    int color[3];
    color[0]=color[1]=color[2]=0;
    if (wd->filter)
      ms_filter_call_method(wd->filter, MS_VIDEO_OUT_GET_BACKGROUND_COLOR, &color);
    yuv_buf_background(fbuf, color[0], color[1], color[2]);
    yuv420p_to_rgb(wd, &wd->fb, wd->rgb);
  }
  return TRUE;
}

static void win_display_update(MSDisplay *obj, int new_image, int new_selfview){
  WinDisplay *wd=(WinDisplay*)obj->data;
  HDC hdc;
  BITMAPINFOHEADER bi;
  RECT rect;
  RECT rect_full;
  bool_t ret;
  int ratiow;
  int ratioh;
  int i;
  int w;
  int h;
  int corner;
  float sv_scalefactor;
  float sv_pos[3];
  int color[3];

  HDC dd_hdc;
  HBITMAP dd_bmp;
  HBRUSH brush;
  BOOL dont_draw;

  if (wd->window==NULL) return;
  hdc=GetDC(wd->window);
  if (hdc==NULL) {
    ms_error("Could not get window dc");
    return;
  }
  if (new_image>0)
    yuv420p_to_rgb(wd, &wd->fb, wd->rgb);
  memset(&bi,0,sizeof(bi));
  bi.biSize=sizeof(bi);
  GetClientRect(wd->window,&rect);
  if (rect.right==0 || rect.bottom==0) {
    ReleaseDC(NULL,hdc);
    return;
  }
  memcpy(&rect_full, &rect, sizeof(rect_full));

  bi.biWidth=wd->fb.w;
  bi.biHeight=wd->fb.h;
  bi.biPlanes=1;
#ifdef USE_RGB32BIT
  bi.biBitCount=32;
#else
  bi.biBitCount=24;
#endif
  bi.biCompression=BI_RGB;
  bi.biSizeImage=wd->rgb_len;

  ratiow=wd->fb.w;
  ratioh=wd->fb.h;

  ms_video_ratio_reduce(&ratiow, &ratioh);

  for (i=0;i<10;i++) {
    w = rect.right/ratiow*ratiow;
    h = rect.bottom/ratioh*ratioh;

    if (w==0 || h==0) {

    }

    if (h*ratiow>w*ratioh)
    {
      w = w;
      h = w*ratioh/ratiow;
    }
    else
    {
      h = h;
      w = h*ratiow/ratioh;
    }

    if (w>rect.right*80/100 || h>rect.bottom*80/100) {
      break;
    }
    ratiow=ratiow/2;
    ratiow=ratiow & ~0x1;
    ratioh=ratioh/2;
    ratioh=ratioh & ~0x1;
    if (ratioh==0 || ratiow==0)
      break;
    ms_video_ratio_reduce(&ratiow, &ratioh);
  }

  if (w <= 1 || h <= 1) {
    ReleaseDC(NULL,hdc);
    return;
  }

  /* if (h*wd->fb.w!=w*wd->fb.h) */
  /*  ms_error("wrong ratio"); */

  dd_hdc = CreateCompatibleDC(hdc);
  if (dd_hdc==NULL) {
    ms_error("Could not get CreateCompatibleDC");
    ReleaseDC(NULL,hdc);
    return;
  }
  dd_bmp = CreateCompatibleBitmap(hdc, rect.right, rect.bottom);
  if (dd_bmp==NULL) {
    ms_error("Could not get CreateCompatibleBitmap");
    ReleaseDC(NULL,hdc);
    return;
  }

  HGDIOBJ old_object = SelectObject(dd_hdc, dd_bmp);

  dont_draw = DrawDibBegin(wd->ddh,dd_hdc, 0, 0, &bi, 0, 0, DDF_BUFFER);

  /* full screen in background color */
  color[0]=color[1]=color[2]=0;
  if (wd->filter)
    ms_filter_call_method(wd->filter, MS_VIDEO_OUT_GET_BACKGROUND_COLOR, &color);

  brush = CreateSolidBrush(RGB(color[0],color[1],color[2]));
  FillRect(dd_hdc, &rect_full, brush); 
  DeleteObject(brush);

  corner = 0;
  sv_scalefactor = SCALE_FACTOR;
  sv_pos[0] = SELVIEW_POS_INACTIVE;
  sv_pos[1] = SELVIEW_POS_INACTIVE;
  sv_pos[2] = 0;
  if (wd->filter)
    ms_filter_call_method(wd->filter, MS_VIDEO_OUT_GET_CORNER, &corner);
  if (wd->filter)
    ms_filter_call_method(wd->filter, MS_VIDEO_OUT_GET_SCALE_FACTOR, &sv_scalefactor);
  if (wd->filter)
    ms_filter_call_method(wd->filter, MS_VIDEO_OUT_GET_SELFVIEW_POS, sv_pos);

  if (wd->rgb_selfview==NULL || corner==-1) {
    ret=DrawDibDraw(wd->ddh,dd_hdc,
      (rect.right-w)/2,
      (rect.bottom-h)/2,
      w,
      h,
      &bi,wd->rgb,
      0,0,bi.biWidth,bi.biHeight,dont_draw?DDF_DONTDRAW:0);

  } else {

    int w_selfview = rect.right - w;
    int h_selfview = rect.bottom - h;

    if ((h_selfview < h/sv_scalefactor && w_selfview < w/sv_scalefactor) || corner<=3 || sv_pos[0]!=SELVIEW_POS_INACTIVE)
    {
      ret=DrawDibDraw(wd->ddh,dd_hdc,
        (rect.right-w)/2,
        (rect.bottom-h)/2,
        w,
        h,
        &bi,wd->rgb,
        0,0,bi.biWidth,bi.biHeight,dont_draw?DDF_DONTDRAW:0);

      //preserve ratio
      ratiow=wd->fb_selfview.w;
      ratioh=wd->fb_selfview.h;
      ms_video_ratio_reduce(&ratiow, &ratioh);

      w_selfview = (int)(w/sv_scalefactor);
      w_selfview = w_selfview/ratiow*ratiow;
      h_selfview = w_selfview*ratioh/ratiow;

      if (rect.right>100 && rect.bottom>100)
      {
        int x_sv;
        int y_sv;
        if (new_selfview>0)
          yuv420p_to_rgb_selfview(wd, &wd->fb_selfview, wd->rgb_selfview);

        //HPEN hpenDot;
        //hpenDot = CreatePen(PS_SOLID, 1, RGB(10, 10, 10));
        //SelectObject(dd_hdc, hpenDot);
        if (sv_pos[0]!=SELVIEW_POS_INACTIVE)
        {
          x_sv = (int)((rect.right*sv_pos[0])/100.0-w_selfview/2);
          y_sv = (int)((rect.bottom*sv_pos[1])/100.0-h_selfview/2);
        }
        else if (corner==1 || corner==4+1)
        {
          /* top left corner */
          x_sv = 20;
          y_sv = 20;
        }
        else if (corner==2 || corner==4+2)
        {
          /* top right corner */
          x_sv = (rect.right-w_selfview-20);
          y_sv = 20;
        }
        else if (corner==3 || corner==4+3)
        {
          /* bottom left corner */
          x_sv = 20;
          y_sv = (rect.bottom-h_selfview-20);
        }
        else /* corner = 0: default */
        {
          /* bottom right corner */
          x_sv = (rect.right-w_selfview-20);
          y_sv = (rect.bottom-h_selfview-20);
        }


        Rectangle(dd_hdc, x_sv-2, y_sv-2, x_sv+w_selfview+2, y_sv+h_selfview+2); 
        ret=DrawDibDraw(wd->ddh,dd_hdc,
          x_sv,
          y_sv,
          w_selfview,
          h_selfview,
          &bi,wd->rgb_selfview,
          0,0,bi.biWidth,bi.biHeight,dont_draw?DDF_DONTDRAW:0);
      }
    }
    else
    {
      //preserve ratio
      ratiow=wd->fb_selfview.w;
      ratioh=wd->fb_selfview.h;
      ms_video_ratio_reduce(&ratiow, &ratioh);

      if (new_selfview>0)
        yuv420p_to_rgb_selfview(wd, &wd->fb_selfview, wd->rgb_selfview);
      if (w_selfview >= w/sv_scalefactor)
      {
        w_selfview = w_selfview/ratiow*ratiow;
        h_selfview = w_selfview*ratioh/ratiow;

        ret=DrawDibDraw(wd->ddh,dd_hdc,
          0,
          (rect.bottom-h)/2,
          w,
          h,
          &bi,wd->rgb,
          0,0,bi.biWidth,bi.biHeight,dont_draw?DDF_DONTDRAW:0);

        Rectangle(dd_hdc,
          (rect.right-w_selfview)-4-2,
          (rect.bottom-h_selfview)/2-2,
          (rect.right)-4+2,
          (rect.bottom+h_selfview)/2+2);

        ret=DrawDibDraw(wd->ddh,dd_hdc,
          (rect.right-w_selfview)-4,
          (rect.bottom-h_selfview)/2,
          w_selfview,
          h_selfview,
          &bi,wd->rgb_selfview,
          0,0,bi.biWidth,bi.biHeight,dont_draw?DDF_DONTDRAW:0);
      }
      else
      {
        h_selfview = h_selfview/ratioh*ratioh;
        w_selfview = h_selfview*ratiow/ratioh;

        ret=DrawDibDraw(wd->ddh,dd_hdc,
          (rect.right-w)/2,
          0,
          w,
          h,
          &bi,wd->rgb,
          0,0,bi.biWidth,bi.biHeight,dont_draw?DDF_DONTDRAW:0);

        Rectangle(dd_hdc,
          (rect.right-w_selfview)/2-2,
          (rect.bottom-h_selfview)-4-2,
          (rect.right+w_selfview)/2+2,
          (rect.bottom)-4+2);

        ret=DrawDibDraw(wd->ddh,dd_hdc,
          (rect.right-w_selfview)/2,
          (rect.bottom-h_selfview)-4,
          w_selfview,
          h_selfview,
          &bi,wd->rgb_selfview,
          0,0,bi.biWidth,bi.biHeight,dont_draw?DDF_DONTDRAW:0);
      }
    }
  }

  DrawDibEnd(wd->ddh);
  BitBlt(hdc, 0, 0, rect.right, rect.bottom, dd_hdc, 0, 0, SRCCOPY);
  SelectObject(dd_hdc, old_object);

  DeleteObject(dd_bmp);
  DeleteDC(dd_hdc);

  wd->last_rect_w=rect.right;
  wd->last_rect_h=rect.bottom;

  if (!ret) ms_error("DrawDibDraw failed.");
  ReleaseDC(NULL,hdc);
}

static void win_display_uninit(MSDisplay *obj){
  WinDisplay *wd=(WinDisplay*)obj->data;
  if (wd==NULL)
    return;
  if (wd->window && !obj->use_external_window) DestroyWindow(wd->window);
  if (wd->ddh) DrawDibClose(wd->ddh);
  if (wd->fb_selfview.planes[0]) ms_free(wd->fb_selfview.planes[0]);
  if (wd->rgb_selfview) ms_free(wd->rgb_selfview);
  if (wd->sws_selfview) ms_video_scalercontext_free(wd->sws_selfview);
  if (wd->fb.planes[0]) ms_free(wd->fb.planes[0]);
  if (wd->rgb) ms_free(wd->rgb);
  if (wd->sws) ms_video_scalercontext_free(wd->sws);
  ms_free(wd);
}

int win_display_pollevent(MSDisplay *d, MSDisplayEvent *ev){
  return -1;
}

MSDisplayDesc ms_win_display_desc={
  "win",
  win_display_init,
  NULL,
  NULL,
  win_display_update,
  win_display_uninit,
  win_display_pollevent
};

#endif

struct _OList *display_tables=NULL;
static int display_table_initialized=0;
static MSDisplayDesc *default_display_desc=NULL;

static void display_module_init() {
#if defined(NOWINDOW_DISPLAY)
  display_tables = o_list_append(display_tables, &ms_nowindow_display_desc);
  if (default_display_desc==NULL)
    default_display_desc = &ms_nowindow_display_desc;
#endif
#if defined(WIN32)
  display_tables = o_list_append(display_tables, &ms_win_display_desc);
  if (default_display_desc==NULL)
    default_display_desc = &ms_win_display_desc;
#endif
#if defined(HAVE_X11_EXTENSIONS_XV_H) && defined(HAVE_X11_XLIB_H)
  display_tables = o_list_append(display_tables, &ms_xv_display_desc);
  if (default_display_desc==NULL)
    default_display_desc = &ms_xv_display_desc;
#endif
#if defined(HAVE_SDL13)
  display_tables = o_list_append(display_tables, &ms_sdl13_display_desc);
  if (default_display_desc==NULL)
    default_display_desc = &ms_sdl13_display_desc;
#endif
#if defined(HAVE_SDL)
  display_tables = o_list_append(display_tables, &ms_sdl_display_desc);
  if (default_display_desc==NULL)
    default_display_desc = &ms_sdl_display_desc;
#endif
}

void display_module_remove_all(void) {
  display_tables = o_list_free(display_tables);
}

MSDisplay *ms_display_new_from_name(char *name){
  OList *elem;

  if (display_table_initialized==0) {
    display_module_init();
    display_table_initialized=1;
  }

	for (elem=display_tables;elem!=NULL;elem=o_list_next(elem)){
		MSDisplayDesc *s=(MSDisplayDesc*) elem->data;
    if (strcasecmp(name, s->name)==0) {
      MSDisplay *obj=(MSDisplay *)ms_new0(MSDisplay,1);
      obj->desc=s;
      obj->data=NULL;
      return obj;
    }
  }
  return NULL;
}

MSDisplay *ms_display_new(MSDisplayDesc *desc){
  MSDisplay *obj=(MSDisplay *)ms_new0(MSDisplay,1);
  obj->desc=desc;
  obj->data=NULL;
  return obj;
}

void ms_display_set_window_id(MSDisplay *d, long id){
  d->window_id=id;
  d->use_external_window=TRUE;
}

void ms_display_destroy(MSDisplay *obj){
  obj->desc->uninit(obj);
  ms_free(obj);
}

void ms_display_desc_set_default(MSDisplayDesc *desc){

  if (display_table_initialized==0) {
    display_module_init();
    display_table_initialized=1;
  }

  default_display_desc=desc;
}

MSDisplayDesc * ms_display_desc_get_default(void){

  if (display_table_initialized==0) {
    display_module_init();
    display_table_initialized=1;
  }

  return default_display_desc;
}

void ms_display_desc_set_default_window_id(MSDisplayDesc *desc, long id){
  desc->default_window_id=id;
}

typedef struct VideoOut
{
  MSPicture fbuf;
  MSPicture fbuf_selfview;
  MSPicture local_pic;
  MSRect local_rect;
  mblk_t *local_msg;
  MSVideoSize prevsize;
  int corner; /*for selfview*/
  float scale_factor; /*for selfview*/
  float sv_posx,sv_posy;
  int background_color[3];
  int window_position[2];

  struct MSScalerContext *sws1;
  struct MSScalerContext *sws2;
  MSDisplay *display;
  bool_t own_display;
  bool_t ready;
  bool_t autofit;
  bool_t mirror;
} VideoOut;

static void set_corner(VideoOut *s, int corner)
{
  s->corner=corner;
  s->local_pic.w=((int)(s->fbuf.w/s->scale_factor)) & ~0x1;
  s->local_pic.h=((int)(s->fbuf.h/s->scale_factor)) & ~0x1;
  s->local_rect.w=s->local_pic.w;
  s->local_rect.h=s->local_pic.h;
  if (corner==1) {
    /* top left corner */
    s->local_rect.x=0;
    s->local_rect.y=0;
  } else if (corner==2) {
    /* top right corner */
    s->local_rect.x=s->fbuf.w-s->local_pic.w;
    s->local_rect.y=0;
  } else if (corner==3) {
    /* bottom left corner */
    s->local_rect.x=0;
    s->local_rect.y=s->fbuf.h-s->local_pic.h;
  } else {
    /* default: bottom right corner */
    /* corner can be set to -1: to disable the self view... */
    s->local_rect.x=s->fbuf.w-s->local_pic.w;
    s->local_rect.y=s->fbuf.h-s->local_pic.h;
  }
  s->fbuf_selfview.w=(s->fbuf.w/1) & ~0x1;
  s->fbuf_selfview.h=(s->fbuf.h/1) & ~0x1;
}

static void re_vsize(VideoOut *s, MSVideoSize *sz){
  ms_message("Windows size set to %ix%i",sz->width,sz->height);
}


static void set_vsize(VideoOut *s, MSVideoSize *sz){
  s->fbuf.w=sz->width & ~0x1;
  s->fbuf.h=sz->height & ~0x1;
  set_corner(s,s->corner);
  ms_message("Video size set to %ix%i",s->fbuf.w,s->fbuf.h);
}

static void video_out_init(MSFilter  *f){
  VideoOut *obj=(VideoOut*)ms_new0(VideoOut,1);
  MSVideoSize def_size;
  def_size.width=MS_VIDEO_SIZE_CIF_W;
  def_size.height=MS_VIDEO_SIZE_CIF_H;
  obj->prevsize.width=0;
  obj->prevsize.height=0;
  obj->local_msg=NULL;
  obj->corner=0;
  obj->scale_factor=SCALE_FACTOR;
  obj->sv_posx=obj->sv_posy=SELVIEW_POS_INACTIVE;
  obj->background_color[0]=obj->background_color[1]=obj->background_color[2]=0;
  obj->window_position[0]=obj->window_position[1];
  obj->sws1=NULL;
  obj->sws2=NULL;
  obj->display=NULL;
  obj->own_display=FALSE;
  obj->ready=FALSE;
  obj->autofit=FALSE;
  obj->mirror=FALSE;
  set_vsize(obj,&def_size);
  f->data=obj;
}


static void video_out_uninit(MSFilter *f){
  VideoOut *obj=(VideoOut*)f->data;
  if (obj->display!=NULL && obj->own_display)
    ms_display_destroy(obj->display);
  if (obj->sws1!=NULL){
    ms_video_scalercontext_free(obj->sws1);
    obj->sws1=NULL;
  }
  if (obj->sws2!=NULL){
    ms_video_scalercontext_free(obj->sws2);
    obj->sws2=NULL;
  }
  if (obj->local_msg!=NULL) {
    freemsg(obj->local_msg);
    obj->local_msg=NULL;
  }
  ms_free(obj);
}

static void video_out_prepare(MSFilter *f){
  VideoOut *obj=(VideoOut*)f->data;
  if (obj->display==NULL){
    if (default_display_desc==NULL){
      ms_error("No default display built in !");
      return;
    }
    obj->display=ms_display_new(default_display_desc);
    obj->own_display=TRUE;
  }
  if (!ms_display_init(obj->display,f,&obj->fbuf,&obj->fbuf_selfview)){
    if (obj->own_display) ms_display_destroy(obj->display);
    obj->display=NULL;
  } else {
    yuv_buf_background(&obj->fbuf, obj->background_color[0], obj->background_color[1], obj->background_color[2]);
  }
  if (obj->sws1!=NULL){
    ms_video_scalercontext_free(obj->sws1);
    obj->sws1=NULL;
  }
  if (obj->sws2!=NULL){
    ms_video_scalercontext_free(obj->sws2);
    obj->sws2=NULL;
  }
  if (obj->local_msg!=NULL) {
    freemsg(obj->local_msg);
    obj->local_msg=NULL;
  }
  set_corner(obj,obj->corner);
  obj->ready=TRUE;
}

static int video_out_handle_resizing(MSFilter *f, void *data){
  /* to be removed */
  return -1;
}

static int _video_out_handle_resizing(MSFilter *f, void *data){
  VideoOut *s=(VideoOut*)f->data;
  MSDisplay *disp=s->display;
  int ret = -1;
  if (disp!=NULL){
    MSDisplayEvent ev;
    ret = ms_display_poll_event(disp,&ev);
    if (ret>0){
      if (ev.evtype==MS_DISPLAY_RESIZE_EVENT){
        MSVideoSize sz;
        sz.width=ev.w;
        sz.height=ev.h;
        ms_filter_lock(f);
        if (s->ready){
          re_vsize(s,&sz);
          s->ready=FALSE;
        }
        ms_filter_unlock(f);
      }
      return 1;
    }
  }
  return ret;
}

static void video_out_preprocess(MSFilter *f){
#ifndef HAVE_SDL13
  video_out_prepare(f);
#endif
}


static void video_out_process(MSFilter *f){
  VideoOut *obj=(VideoOut*)f->data;
  mblk_t *inm;
  int update=0;
  int update_selfview=0;
  int i;

  for(i=0;i<100;++i){
    int ret = _video_out_handle_resizing(f, NULL);
    if (ret<0)
      break;
  }
  ms_filter_lock(f);
  if (!obj->ready) video_out_prepare(f);
  if (obj->display==NULL){
    ms_filter_unlock(f);
    if (f->inputs[0]!=NULL)
      ms_queue_flush(f->inputs[0]);
    if (f->inputs[1]!=NULL)
      ms_queue_flush(f->inputs[1]);
    return;
  }

  /* remove all MS_H264 data */
  while (f->inputs[1]!=NULL && (inm=ms_queue_peek_last(f->inputs[1]))!=0) {
    if (mblk_get_video_format(inm)==MS_H264) //pass-through
      {
	ms_queue_remove(f->inputs[1],inm);
	freemsg(inm);
	continue;
      }
    break;
  }
  /* remove all MS_H264 data */
  while (f->inputs[0]!=NULL && (inm=ms_queue_peek_last(f->inputs[0]))!=0) {
    if (mblk_get_video_format(inm)==MS_H264) //pass-through
      {
	ms_queue_remove(f->inputs[0],inm);
	freemsg(inm);
	continue;
      }
    break;
  }

  /*get most recent message and draw it*/
  if (f->inputs[1]!=NULL && (inm=ms_queue_peek_last(f->inputs[1]))!=0) {

    if (obj->corner==-1){
      if (obj->local_msg!=NULL) {
        freemsg(obj->local_msg);
        obj->local_msg=NULL;
      }
    }else if (obj->fbuf_selfview.planes[0]!=NULL) {
      MSPicture src;
      if (yuv_buf_init_from_mblk(&src,inm)==0){

        if (obj->sws2==NULL){
          obj->sws2=ms_video_scalercontext_init(src.w,src.h,MS_YUV420P,
            obj->fbuf_selfview.w,obj->fbuf_selfview.h,MS_YUV420P,
            MS_YUVFAST, NULL, NULL, NULL);
        }
        ms_display_lock(obj->display);
        if (ms_video_scalercontext_convert(obj->sws2,src.planes,src.strides, 0,
          src.h, obj->fbuf_selfview.planes, obj->fbuf_selfview.strides)<0){
            ms_error("Error in ms_video_scalercontext_convert().");
        }
        if (!mblk_get_precious_flag(inm)) ms_yuv_buf_mirror(&obj->fbuf_selfview);
        ms_display_unlock(obj->display);
        update_selfview=1;
      }
    }else{
      MSPicture src;
      if (yuv_buf_init_from_mblk(&src,inm)==0){

        if (obj->sws2==NULL){
          obj->sws2=ms_video_scalercontext_init(src.w,src.h,MS_YUV420P,
            obj->local_pic.w,obj->local_pic.h,MS_YUV420P,
            MS_YUVFAST, NULL, NULL, NULL);
        }
        if (obj->local_msg==NULL){
          obj->local_msg=yuv_buf_alloc(&obj->local_pic,
            obj->local_pic.w,obj->local_pic.h);
        }
        if (obj->local_pic.planes[0]!=NULL)
        {
          if (ms_video_scalercontext_convert(obj->sws2,src.planes,src.strides, 0,
            src.h, obj->local_pic.planes, obj->local_pic.strides)<0){
              ms_error("Error in ms_video_scalercontext_convert().");
          }
          if (!mblk_get_precious_flag(inm)) ms_yuv_buf_mirror(&obj->local_pic);
          update=1;
        }
      }
    }
    ms_queue_flush(f->inputs[1]);
  }

  if (f->inputs[0]!=NULL && (inm=ms_queue_peek_last(f->inputs[0]))!=0) {
    MSPicture src;
    if (yuv_buf_init_from_mblk(&src,inm)==0){
      MSPicture dest;
      int w,h,x,y;
      int ratiow, ratioh;
      int wtmp,htmp;

      /* live change of incoming stream size */
      if (obj->prevsize.width!=src.w || obj->prevsize.height!=src.h)
      {
#if defined(HAVE_X11_EXTENSIONS_XV_H) && defined(HAVE_X11_XLIB_H)
#else
	MSVideoSize sz;
	sz.width=src.w;
	sz.height=src.h;
	set_vsize(obj, &sz);
#endif
	ms_error("incoming data size has changed!");
        video_out_prepare(f);
      }

      obj->prevsize.width=src.w;
      obj->prevsize.height=src.h;
      /* end of live change of incoming stream size */

      /* we want to maintain aspect ratio */
      ratiow=src.w;
      ratioh=src.h;
      ms_video_ratio_reduce(&ratiow, &ratioh);
      wtmp = obj->fbuf.w/ratiow*ratiow;
      htmp = obj->fbuf.h/ratioh*ratioh;

      if (htmp*ratiow>wtmp*ratioh)
      {
        wtmp = wtmp;
        htmp = wtmp*ratioh/ratiow;
      }
      else
      {
        htmp = htmp;
        wtmp = htmp*ratiow/ratioh;
      }

      w=wtmp;
      h=htmp;

      x=((int)(((float)obj->fbuf.w*0.5)-((float)w/2))) & ~0x1;
      y=((int)(((float)obj->fbuf.h*0.5)-((float)h/2))) & ~0x1;

      dest.w=w;
      dest.h=h;
      dest.planes[0]=obj->fbuf.planes[0]+(y*obj->fbuf.strides[0])+x;
      dest.planes[1]=obj->fbuf.planes[1]+((y/2)*obj->fbuf.strides[1])+(x/2);
      dest.planes[2]=obj->fbuf.planes[2]+((y/2)*obj->fbuf.strides[2])+(x/2);
      dest.planes[3]=NULL;
      dest.strides[0]=w;
      dest.strides[1]=w/2;
      dest.strides[2]=w/2;
      dest.strides[3]=0;

      if (obj->sws1==NULL){
        yuv_buf_background(&obj->fbuf, obj->background_color[0], obj->background_color[1], obj->background_color[2]);
        obj->sws1=ms_video_scalercontext_init(src.w,src.h,MS_YUV420P,
          dest.w,dest.h,MS_YUV420P,
          MS_YUVFAST, NULL, NULL, NULL);
      }

      ms_display_lock(obj->display);
      if (ms_video_scalercontext_convert(obj->sws1,src.planes,src.strides, 0,
        src.h, dest.planes, obj->fbuf.strides)<0){
          ms_error("Error in ms_video_scalercontext_convert().");
      }
      if (obj->mirror && !mblk_get_precious_flag(inm)) ms_yuv_buf_mirror(&obj->fbuf);
      ms_display_unlock(obj->display);
    }
    update=1;
    ms_queue_flush(f->inputs[0]);
  }

  /*copy resized local view into main buffer, at bottom left corner:*/
  if (obj->local_msg!=NULL){
    MSPicture corner=obj->fbuf;
    MSVideoSize roi;
    roi.width=obj->local_pic.w;
    roi.height=obj->local_pic.h;
    corner.w=obj->local_pic.w;
    corner.h=obj->local_pic.h;
    corner.planes[0]+=obj->local_rect.x+(obj->local_rect.y*corner.strides[0]);
    corner.planes[1]+=(obj->local_rect.x/2)+((obj->local_rect.y/2)*corner.strides[1]);
    corner.planes[2]+=(obj->local_rect.x/2)+((obj->local_rect.y/2)*corner.strides[2]);
    corner.planes[3]=0;
    ms_display_lock(obj->display);
    ms_yuv_buf_copy(obj->local_pic.planes,obj->local_pic.strides,
      corner.planes,corner.strides,roi);
    ms_display_unlock(obj->display);
  }

  ms_display_update(obj->display, update, update_selfview);
  ms_filter_unlock(f);
}

static int video_out_set_vsize(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  ms_filter_lock(f);
  set_vsize(s,(MSVideoSize*)arg);
#if defined(HAVE_X11_EXTENSIONS_XV_H) && defined(HAVE_X11_XLIB_H)
  s->ready=FALSE;
#endif
  ms_filter_unlock(f);
  return 0;
}

static int video_out_set_display(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  s->display=(MSDisplay*)arg;
  return 0;
}

static int video_out_auto_fit(MSFilter *f, void *arg){
  VideoOut *s=(VideoOut*)f->data;
  s->autofit=*(int*)arg;
  return 0;
}

static int video_out_set_corner(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  s->sv_posx=s->sv_posy=SELVIEW_POS_INACTIVE;
  ms_filter_lock(f);
  set_corner(s, *(int*)arg);
  
  if (s->ready==TRUE && s->display){
    ms_display_lock(s->display);
    {
      int w=s->fbuf.w;
      int h=s->fbuf.h;
      int ysize=w*h;
      int usize=ysize/4;

      memset(s->fbuf.planes[0], 0, ysize);
      memset(s->fbuf.planes[1], 0, usize);
      memset(s->fbuf.planes[2], 0, usize);
      s->fbuf.planes[3]=NULL;
    }
    ms_display_unlock(s->display);
  }
  ms_filter_unlock(f);
  return 0;
}

static int video_out_get_corner(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  *((int*)arg)=s->corner;
  return 0;
}

static int video_out_set_scalefactor(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  s->scale_factor = *(float*)arg;
  if (s->scale_factor<0.5f)
    s->scale_factor = 0.5f;
  ms_filter_lock(f);
  set_corner(s, s->corner);
  if (s->display && s->ready==TRUE){
    ms_display_lock(s->display);
    {
      int w=s->fbuf.w;
      int h=s->fbuf.h;
      int ysize=w*h;
      int usize=ysize/4;

      memset(s->fbuf.planes[0], 0, ysize);
      memset(s->fbuf.planes[1], 0, usize);
      memset(s->fbuf.planes[2], 0, usize);
      s->fbuf.planes[3]=NULL;
    }
    ms_display_unlock(s->display);
  }
  ms_filter_unlock(f);
  return 0;
}

static int video_out_get_scalefactor(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  *((float*)arg)=(float)s->scale_factor;
  return 0;
}


static int video_out_enable_mirroring(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  s->mirror=*(int*)arg;
  return 0;
}

static int video_out_get_native_window_id(MSFilter *f, void*arg){
  VideoOut *s=(VideoOut*)f->data;
  unsigned long *id=(unsigned long*)arg;
  *id=0;
  if (s->display){
    *id=s->display->window_id;
    return 0;
  }
  return -1;
}

static int video_out_set_selfview_pos(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  s->sv_posx=((float*)arg)[0];
  s->sv_posy=((float*)arg)[1];
  s->scale_factor=(float)100.0/((float*)arg)[2];
  return 0;
}

static int video_out_get_selfview_pos(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  ((float*)arg)[0]=s->sv_posx;
  ((float*)arg)[1]=s->sv_posy;
  ((float*)arg)[2]=(float)100.0/s->scale_factor;
  return 0;
}

static int video_out_set_background_color(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  s->background_color[0]=((int*)arg)[0];
  s->background_color[1]=((int*)arg)[1];
  s->background_color[2]=((int*)arg)[2];
  return 0;
}

static int video_out_get_background_color(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  ((int*)arg)[0]=s->background_color[0];
  ((int*)arg)[1]=s->background_color[1];
  ((int*)arg)[2]=s->background_color[2];
  return 0;
}

static int video_out_set_window_position(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  s->window_position[0]=((int*)arg)[0];
  s->window_position[1]=((int*)arg)[1];
  return 0;
}

static int video_out_get_window_position(MSFilter *f,void *arg){
  VideoOut *s=(VideoOut*)f->data;
  ((int*)arg)[0]=s->window_position[0];
  ((int*)arg)[1]=s->window_position[1];
  return 0;
}

static int get_statistics(MSFilter *f, void *arg){
  int i;
  ms_warning("filter: %s[%i->%i]", f->desc->name,
    f->desc->ninputs, f->desc->noutputs);
  for (i=0;i<f->desc->ninputs;i++) {
    if (f->inputs[i]!=NULL) ms_warning("filter: %s in[%i]=%i", f->desc->name,
      i, f->inputs[i]->q.q_mcount);
  }
  for (i=0;i<f->desc->noutputs;i++) {
    if (f->outputs[i]!=NULL) ms_warning("filter: %s out[%i]=%i", f->desc->name,
      i, f->outputs[i]->q.q_mcount);
  }
  return 0;
}

static MSFilterMethod methods[]={
  {	MS_FILTER_SET_VIDEO_SIZE	,	video_out_set_vsize },
  {	MS_VIDEO_OUT_SET_DISPLAY	,	video_out_set_display},
  {	MS_VIDEO_OUT_SET_CORNER 	,	video_out_set_corner},
  {	MS_VIDEO_OUT_AUTO_FIT		,	video_out_auto_fit},
  {	MS_VIDEO_OUT_HANDLE_RESIZING	,	video_out_handle_resizing},
  {	MS_VIDEO_OUT_ENABLE_MIRRORING	,	video_out_enable_mirroring},
  {	MS_VIDEO_OUT_GET_NATIVE_WINDOW_ID,	video_out_get_native_window_id},
  {	MS_VIDEO_OUT_GET_CORNER 	,	video_out_get_corner},
  {	MS_VIDEO_OUT_SET_SCALE_FACTOR 	,	video_out_set_scalefactor},
  {	MS_VIDEO_OUT_GET_SCALE_FACTOR 	,	video_out_get_scalefactor},
  {	MS_VIDEO_OUT_SET_SELFVIEW_POS 	 ,	video_out_set_selfview_pos},
  {	MS_VIDEO_OUT_GET_SELFVIEW_POS    ,  video_out_get_selfview_pos},
  {	MS_VIDEO_OUT_SET_BACKGROUND_COLOR    ,  video_out_set_background_color},
  {	MS_VIDEO_OUT_GET_BACKGROUND_COLOR    ,  video_out_get_background_color},
  {	MS_VIDEO_OUT_SET_WINDOW_POSITION    ,  video_out_set_window_position},
  {	MS_VIDEO_OUT_GET_WINDOW_POSITION    ,  video_out_get_window_position},
  /* methods for compatibility with the MSVideoDisplay interface*/
  {	MS_VIDEO_DISPLAY_SET_LOCAL_VIEW_MODE , video_out_set_corner },
  {	MS_VIDEO_DISPLAY_ENABLE_AUTOFIT			, video_out_auto_fit },
  {	MS_VIDEO_DISPLAY_ENABLE_MIRRORING		, video_out_enable_mirroring },
  {	MS_VIDEO_DISPLAY_GET_NATIVE_WINDOW_ID	, video_out_get_native_window_id },
  {	MS_VIDEO_DISPLAY_SET_LOCAL_VIEW_SCALEFACTOR	, video_out_set_scalefactor },
  {	MS_VIDEO_DISPLAY_SET_SELFVIEW_POS 	 ,	video_out_set_selfview_pos},
  {	MS_VIDEO_DISPLAY_GET_SELFVIEW_POS    ,  video_out_get_selfview_pos},
  {	MS_VIDEO_DISPLAY_SET_BACKGROUND_COLOR    ,  video_out_set_background_color},
  {	MS_FILTER_GET_STATISTICS, get_statistics },

  {	0	,NULL}
};

#ifdef _MSC_VER

MSFilterDesc ms_video_out_desc={
  MS_VIDEO_OUT_ID,
  "MSVideoOut",
  N_("A generic video display"),
  MS_FILTER_OTHER,
  NULL,
  2,
  0,
  video_out_init,
  video_out_preprocess,
  video_out_process,
  NULL,
  video_out_uninit,
  methods
};

#else

MSFilterDesc ms_video_out_desc={
  .id=MS_VIDEO_OUT_ID,
  .name="MSVideoOut",
  .text=N_("A generic video display"),
  .category=MS_FILTER_OTHER,
  .ninputs=2,
  .noutputs=0,
  .init=video_out_init,
  .preprocess=video_out_preprocess,
  .process=video_out_process,
  .uninit=video_out_uninit,
  .methods=methods
};

#endif

MS_FILTER_DESC_EXPORT(ms_video_out_desc)

#else

MSDisplay *ms_display_new(MSDisplayDesc *desc){
  return NULL; //not implemented
}

void ms_display_set_window_id(MSDisplay *d, long id){
  //not implemented
}

void ms_display_destroy(MSDisplay *obj){
  //not implemented
}

void ms_display_desc_set_default(MSDisplayDesc *desc){
  //not implemented
}

MSDisplayDesc * ms_display_desc_get_default(void){
  return NULL; //not implemented
}

#endif
